#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
autosign: Auto-sign MMGen transactions, message files and XMR wallet output files
"""

import sys

from .util import msg, die, fmt_list, exit_if_mswin, async_run

exit_if_mswin('autosigning')

opts_data = {
	'sets': [('stealth_led', True, 'led', True)],
	'text': {
		'desc': 'Auto-sign MMGen transactions, message files and XMR wallet output files',
		'usage':'[opts] [operation]',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long (global) options
-c, --coins=c         Coins to sign for (comma-separated list)
-I, --no-insert-check Don’t check for device insertion
-k, --keys-from-file=F Use wif keys listed in file ‘F’ for signing non-MMGen
                      inputs. The file may be MMGen encrypted if desired. The
                      ‘setup’ operation creates a temporary encrypted copy of
                      the file in volatile memory for use during the signing
                      session, thus permitting the deletion of the original
                      file for increased security.
-l, --seed-len=N      Specify wallet seed length of ‘N’ bits (for setup only)
-L, --led             Use status LED to signal standby, busy and error
-m, --mountpoint=M    Specify an alternate mountpoint 'M'
                      (default: {asi.dfl_mountpoint!r})
-M, --mnemonic-fmt=F  During setup, prompt for mnemonic seed phrase of format
                      'F' (choices: {mn_fmts}; default: {asi.dfl_mn_fmt!r})
-n, --no-summary      Don’t print a transaction summary
-r, --macos-ramdisk-size=S  Set the size (in MB) of the ramdisk used to store
                      the offline signing wallet(s) on macOS machines.  By
                      default, a runtime-calculated value will be used. This
                      option is of interest only for setups with unusually
                      large Monero wallets
-s, --stealth-led     Stealth LED mode - signal busy and error only, and only
                      after successful authorization.
-S, --full-summary    Print a full summary of each signed transaction after
                      each autosign run. The default list of non-MMGen outputs
                      will not be printed.
-q, --quiet           Produce quieter output
-v, --verbose         Produce more verbose output
-w, --wallet-dir=D    Specify an alternate wallet dir
                      (default: {asi.dfl_wallet_dir!r})
-W, --allow-non-wallet-swap Allow signing of swap transactions that send funds
                      to non-wallet addresses
-x, --xmrwallets=L    Range or list of wallets to be used for XMR autosigning
""",
	'notes': """

                               OPERATIONS

clean     - clean the removable device of unneeded files, removing only non-
            essential data
gen_key   - generate the wallet encryption key and copy it to the removable
            device mounted at mountpoint ‘{asi.mountpoint}’ (as currently
            configured)
setup     - full setup: run ‘gen_key’ and create temporary signing wallet(s)
            for all configured coins
xmr_setup - set up Monero temporary signing wallet(s).  Not required during
            normal operation: use ‘setup’ with --xmrwallets instead
macos_ramdisk_setup - set up the ramdisk used for storing the temporary signing
            wallet(s) (macOS only).  Required only when creating the wallet(s)
            manually, without ‘setup’
macos_ramdisk_delete - delete the macOS ramdisk
disable_swap - disable disk swap to prevent potentially sensitive data in
            volatile memory from being swapped to disk.  Applicable only when
            creating temporary signing wallet(s) manually, without ‘setup’
enable_swap - reenable disk swap.  For testing only, should not be invoked in
            a production environment
wait      - start in loop mode: wait-mount-sign-unmount-wait
wipe_key  - wipe the wallet encryption key on the removable device, making
            signing transactions or stealing the user’s seed impossible.
            The operation is intended as a ‘kill switch’ and thus performed
            without prompting


                               USAGE NOTES

If no operation is specified, this program mounts a removable device
(typically a USB flash drive) containing unsigned MMGen transactions, message
files, and/or XMR wallet output files, signs them, unmounts the removable
device and exits.

If invoked with ‘wait’, the program waits in a loop, mounting the removable
device, performing signing operations and unmounting the device every time it
is inserted.

On supported platforms (currently Orange Pi, Rock Pi and Raspberry Pi boards),
the status LED indicates whether the program is busy or in standby mode, i.e.
ready for device insertion or removal.

The removable device must have a partition with a filesystem labeled MMGEN_TX
and a user-writable root directory.  For interoperability between OS-es, it’s
recommended to use the exFAT file system.

On both the signing and online machines the mountpoint ‘{asi.mountpoint}’
(as currently configured) must exist.  Linux (not macOS) machines must have
an ‘/etc/fstab’ with the following entry:

    LABEL=MMGEN_TX {asi.mountpoint} auto noauto,user 0 0

Signing is performed with a temporary wallet created in volatile memory in
the directory ‘{asi.wallet_dir}’ (as currently configured).  The wallet is
encrypted with a 32-byte password saved in the file ‘autosign.key’ in the
root of the removable device’s filesystem.

The password and temporary wallet may be created in one operation by invoking
‘mmgen-autosign setup’ with the removable device inserted.  In this case, the
temporary wallet is created from the user’s default wallet, if it exists and
the user so desires.  If not, the user is prompted to enter a seed phrase.

Alternatively, the password and temporary wallet may be created separately by
first invoking ‘mmgen-autosign gen_key’ and then creating and encrypting the
wallet using the -P (--passwd-file) option:

    $ mmgen-walletconv -iwords -d{asi.wallet_dir} -p1 -N -P{asi.mountpoint}/autosign.key -Lfoo

Note that the hash preset must be ‘1’.  To use a wallet file as the source
instead of an MMGen seed phrase, omit the ‘-i’ option and add the wallet
file path to the end of the command line.  Multiple temporary wallets may
be created in this way and used for signing (note, however, that for XMR
operations only one wallet is supported).

Autosigning is currently supported on Linux and macOS only.


                               SECURITY NOTE

By placing wallet and password on separate devices, this program creates
a two-factor authentication setup whereby an attacker must gain physical
control of both the removable device and signing machine in order to sign
transactions.  It’s therefore recommended to always keep the removable device
secure, separated from the signing machine and hidden (in your pocket, for
example) when not transacting.  In addition, since login access on the
signing machine is required to steal the user’s seed, it’s good practice
to lock the signing machine’s screen once the setup process is complete.

As a last resort, cutting power to the signing machine will destroy the
volatile memory where the temporary wallet resides and foil any attack,
even if you’ve lost control of the removable device.

Always remember to power off the signing machine when your signing session
is over.
"""
	},
	'code': {
		'options': lambda s: s.format(
			asi     = asi,
			mn_fmts = fmt_list(asi.mn_fmts, fmt='no_spc'),
		),
		'notes': lambda s: s.format(asi=asi)
	}
}

def main(do_loop):

	asi.init_led()
	asi.init_exit_handler()

	async def do():
		await asi.check_daemons_running()
		if do_loop:
			await asi.main_loop()
		else:
			ret = await asi.do_sign()
			asi.at_exit(not ret)

	async_run(do())

from .cfg import Config
from .autosign import Autosign

cfg = Config(
	opts_data = opts_data,
	init_opts = {
		'out_fmt': 'wallet',
		'usr_randchars': 0,
		'hash_preset': '1',
		'label': 'Autosign Wallet',
	},
	caller_post_init = True)

cmd = cfg._args[0] if len(cfg._args) == 1 else 'sign' if not cfg._args else cfg._usage()

if cmd not in Autosign.cmds + Autosign.util_cmds:
	die(1, f'‘{cmd}’: unrecognized command')

if cmd != 'setup':
	for opt in ('seed_len', 'mnemonic_fmt', 'keys_from_file'):
		if getattr(cfg, opt):
			die(1, f'--{opt.replace("_", "-")} is valid only for the ‘setup’ operation')

if cmd not in ('sign', 'wait'):
	for opt in ('no_summary', 'led', 'stealth_led', 'full_summary'):
		if getattr(cfg, opt):
			die(1, f'--{opt.replace("_", "-")} is not valid for the ‘{cmd}’ operation')

asi = Autosign(cfg, cmd=cmd)

cfg._post_init()

if cmd == 'gen_key':
	asi.gen_key()
elif cmd == 'setup':
	asi.setup()
	from .ui import keypress_confirm
	if cfg.xmrwallets and keypress_confirm(cfg, '\nContinue with Monero setup?', default_yes=True):
		msg('')
		asi.xmr_setup()
	asi.do_umount()
elif cmd == 'xmr_setup':
	if not cfg.xmrwallets:
		die(1, 'Please specify a wallet or range of wallets with the --xmrwallets option')
	asi.do_mount()
	asi.xmr_setup()
	asi.do_umount()
elif cmd.startswith('macos_ramdisk'):
	if sys.platform != 'darwin':
		die(1, f'The ‘{cmd}’ operation is for the macOS platform only')
	getattr(asi, cmd)()
elif cmd == 'enable_swap':
	asi.swap.enable()
elif cmd == 'disable_swap':
	asi.swap.disable()
elif cmd == 'sign':
	main(do_loop=False)
elif cmd == 'wait':
	main(do_loop=True)
elif cmd == 'clean':
	asi.do_mount()
	asi.clean_old_files()
	asi.do_umount()
elif cmd == 'wipe_key':
	asi.do_mount()
	asi.wipe_encryption_key()
	asi.do_umount()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.common: Base class and shared routines for the 'mmgen-tool' utility
"""

from ..objmethods import MMGenObject

def options_annot_str(l):
	return "(valid choices: '{}')".format("','".join(l))

class tool_cmd_base(MMGenObject):

	need_proto = False
	need_addrtype = False
	need_amt = False

	def __init__(self, cfg, *, cmdname=None, proto=None, mmtype=None):

		self.cfg = cfg

		if self.need_proto:
			from ..protocol import init_proto_from_cfg
			self.proto = proto or cfg._proto or init_proto_from_cfg(cfg, need_amt=True)

		if self.need_addrtype:
			from ..addr import MMGenAddrType
			self.mmtype = MMGenAddrType(
				self.proto,
				mmtype or cfg.type or self.proto.dfl_mmtype)

	@property
	def user_commands(self):
		return {k:v for k, v in type(self).__dict__.items() if callable(v) and not k.startswith('_')}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.addrdata: Bitcoin base protocol addrdata classes
"""

from ...addrdata import TwAddrData

class BitcoinTwAddrData(TwAddrData):

	msgs = {
		'multiple_acct_addrs': """
			ERROR: More than one address found for account: {acct!r}.
			Your 'wallet.dat' file appears to have been altered by a non-{proj} program.
			Please restore your tracking wallet from a backup or create a new one and
			re-import your addresses.
		"""
	}

	async def get_tw_data(self, *, twctl=None):
		self.cfg._util.vmsg('Getting address data from tracking wallet')
		c = self.rpc
		if 'label_api' in c.caps:
			accts = await c.call('listlabels')
			ll = await c.batch_call('getaddressesbylabel', [(k,) for k in accts])
			alists = [list(a.keys()) for a in ll]
		else:
			accts = await c.call('listaccounts', 0, True)
			alists = await c.batch_call('getaddressesbyaccount', [(k,) for k in accts])
		return list(zip(accts, alists))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
protocol: Coin protocol base classes and initializer
"""

from collections import namedtuple

from .cfg import gc
from .base_obj import Lockable
from .objmethods import MMGenObject

decoded_wif = namedtuple('decoded_wif', ['sec', 'pubkey_type', 'compressed'])
decoded_addr = namedtuple('decoded_addr', ['bytes', 'ver_bytes', 'fmt'])
decoded_addr_multiview = namedtuple('mv_decoded_addr', ['bytes', 'ver_bytes', 'fmt', 'addr', 'views', 'view_pref'])
parsed_addr = namedtuple('parsed_addr', ['ver_bytes', 'data'])

_finfo = namedtuple('fork_info', ['height', 'hash', 'name', 'replayable'])
_nw = namedtuple('coin_networks', ['mainnet', 'testnet', 'regtest'])

class CoinProtocol(MMGenObject):

	proto_info = namedtuple('proto_info', ['name', 'trust_level']) # trust levels: see altcoin/params.py

	# keys are mirrored in gc.core_coins:
	coins = {
		'btc': proto_info('Bitcoin',         5),
		'bch': proto_info('BitcoinCash',     5),
		'ltc': proto_info('Litecoin',        5),
		'eth': proto_info('Ethereum',        4),
		'etc': proto_info('EthereumClassic', 4),
		'zec': proto_info('Zcash',           2),
		'xmr': proto_info('Monero',          5),
		'rune': proto_info('THORChain',      4)
	}

	class Base(Lockable):
		base_proto = None
		base_proto_coin  = None
		base_coin  = None
		is_fork_of = None
		chain_names = None
		is_vm = False
		is_evm = False
		has_usr_fee = True
		rpc_type = 'local'
		networks   = ('mainnet', 'testnet', 'regtest')
		decimal_prec = 28
		_set_ok = ('tokensym',)

		def __init__(self, cfg, coin, *, name, network, tokensym=None, need_amt=False):
			self.cfg        = cfg
			self.coin       = coin.upper()
			self.coin_id    = self.coin
			self.name       = name
			self.network    = network
			self.tokensym   = tokensym
			self.cls_name   = type(self).__name__
			self.testnet    = network in ('testnet', 'regtest')
			self.regtest    = network == 'regtest'
			self.networks   = tuple(k for k, v in self.network_names._asdict().items() if v)
			self.network_id = coin.lower() + {
				'mainnet': '',
				'testnet': '_tn',
				'regtest': '_rt',
			}[network]

			if hasattr(self, 'wif_ver_num'):
				self.wif_ver_bytes = {k:bytes.fromhex(v) for k, v in self.wif_ver_num.items()}
				self.wif_ver_bytes_to_pubkey_type = {v:k for k, v in self.wif_ver_bytes.items()}
				vbs = list(self.wif_ver_bytes.values())
				self.wif_ver_bytes_len = len(vbs[0]) if len(set(len(b) for b in vbs)) == 1 else None

			if hasattr(self, 'addr_ver_info'):
				self.addr_ver_bytes = {bytes.fromhex(k):v for k, v in self.addr_ver_info.items()}
				self.addr_fmt_to_ver_bytes = {v:k for k, v in self.addr_ver_bytes.items()}
				self.addr_ver_bytes_len = len(list(self.addr_ver_bytes)[0])

			if gc.cmd_caps:
				for cap in gc.cmd_caps.caps:
					if cap not in self.mmcaps:
						from .util import die
						die(2, f'Command {gc.prog_name!r} not supported for coin {self.coin}')

			if self.chain_names:
				self.chain_name = self.chain_names[0] # first chain name is default
			else:
				self.chain_names = [self.network]
				self.chain_name = self.network

			if self.tokensym:
				assert self.name.startswith('Ethereum'), 'CoinProtocol.Base_chk1'

			if self.base_coin in ('ETH', 'XMR'):
				from .util2 import get_keccak
				self.keccak_256 = get_keccak(cfg)

			if need_amt:
				from . import amt
				from decimal import getcontext
				self.coin_amt = getattr(amt, self.coin_amt)
				self.max_tx_fee = self.coin_amt(str(self.max_tx_fee)) if hasattr(self, 'max_tx_fee') else None
				getcontext().prec = self.decimal_prec
			else:
				self.coin_amt = None
				self.max_tx_fee = None

			self.set_cfg_opts()

		def set_cfg_opts(self):
			pass

		@property
		def dcoin(self):
			return self.coin

		@classmethod
		def chain_name_to_network(cls, cfg, coin, chain_name):
			"""
			The generic networks 'mainnet', 'testnet' and 'regtest' are required for all coins
			that support transaction operations.

			For protocols that have specific names for chains corresponding to these networks,
			the attribute 'chain_name' is used, while 'network' retains the generic name.
			For Bitcoin and Bitcoin forks, 'network' and 'chain_name' are equivalent.
			"""
			for network in ('mainnet', 'testnet', 'regtest'):
				proto = init_proto(cfg, coin, network=network)
				for proto_chain_name in proto.chain_names:
					if chain_name == proto_chain_name:
						return network
			raise ValueError(f'{chain_name}: unrecognized chain name for coin {coin}')

		@staticmethod
		def parse_network_id(network_id):
			nid = namedtuple('parsed_network_id', ['coin', 'network'])
			if network_id.endswith('_tn'):
				return nid(network_id[:-3], 'testnet')
			elif network_id.endswith('_rt'):
				return nid(network_id[:-3], 'regtest')
			else:
				return nid(network_id, 'mainnet')

		@staticmethod
		def create_network_id(coin, network):
			return coin.lower() + {'mainnet':'', 'testnet':'_tn', 'regtest':'_rt'}[network]

		def cap(self, s):
			return s in self.caps

		def get_addr_len(self, addr_fmt):
			return self.addr_len

		def decode_addr_bytes(self, addr_bytes):
			vlen = self.addr_ver_bytes_len
			return decoded_addr(
				addr_bytes[vlen:],
				addr_bytes[:vlen],
				self.addr_ver_bytes[addr_bytes[:vlen]])

		def coin_addr(self, addr):
			from .addr import CoinAddr
			return CoinAddr(proto=self, addr=addr)

		def addr_type(self, id_str):
			from .addr import MMGenAddrType
			return MMGenAddrType(proto=self, id_str=id_str)

		def viewkey(self, viewkey_str):
			raise NotImplementedError(f'{self.name} protocol does not support view keys')

		def base_proto_subclass(self, cls, modname, *, sub_clsname=None, is_token=False):
			"""
			magic module loading and class selection
			"""
			modpath = f'mmgen.proto.{self.base_proto_coin.lower()}.{modname}'

			clsname = (
				self.mod_clsname
				+ ('Token' if self.tokensym or is_token else '')
				+ cls.__name__)

			import importlib
			if sub_clsname:
				return getattr(getattr(importlib.import_module(modpath), clsname), sub_clsname)
			else:
				return getattr(importlib.import_module(modpath), clsname)

	class RPC:

		# prefixed with coin, e.g. ‘ltc_rpc_host’: refvals taken from proto class
		coin_cfg_opts = ()

		# prefixed with coin + network, e.g. ‘eth_mainnet_chain_names’: refvals taken from proto class
		proto_cfg_opts = ()

		# default vals (refvals): bool(val) must be False (val = None -> option takes no parameter)
		ignore_daemon_version = None
		rpc_host              = ''
		rpc_port              = 0
		rpc_user              = ''
		rpc_password          = ''
		tw_name               = ''
		daemon_id             = ''

		@classmethod
		def get_opt_clsval(cls, cfg, opt):
			coin, *rem = opt.split('_', 2)
			network = rem[0] if rem[0] in init_proto(cfg, coin, return_cls=True).network_names else None
			opt_name = '_'.join(rem[bool(network):])
			if ((network is None and opt_name in cls.coin_cfg_opts) or
				(network and opt_name in cls.proto_cfg_opts)):
				# raises AttributeError on failure:
				return getattr(init_proto(cfg, coin, network=network, return_cls=True), opt_name)
			else:
				raise AttributeError(f'{opt_name}: unrecognized attribute')

		def set_cfg_opts(self):
			for opt in self.cfg.__dict__:
				if opt.startswith(self.coin.lower() + '_'):
					res = opt.split('_', 2)[1:]
					network = res[0] if res[0] in self.network_names else None
					if network is None or network == self.network:
						setattr(self, '_'.join(res[bool(network):]), getattr(self.cfg, opt))

	class Secp256k1(RPC, Base):
		"""
		Bitcoin and Ethereum protocols inherit from this class
		"""
		secp256k1_group_order = 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141
		privkey_len  = 32
		pubkey_types = ('std',)

		def parse_addr(self, ver_bytes, addr_bytes, fmt):
			return parsed_addr(
				ver_bytes  = ver_bytes,
				data       = addr_bytes,
			)

		def preprocess_key(self, sec, pubkey_type):
			# Key must be non-zero and less than group order of secp256k1 curve
			if 0 < int.from_bytes(sec, 'big') < self.secp256k1_group_order:
				return sec
			else: # chance of this is less than 1 in 2^127
				from .util import die, ymsg
				pk = int.from_bytes(sec, 'big')
				if pk == 0: # chance of this is 1 in 2^256
					die(4, 'Private key is zero!')
				elif pk == self.secp256k1_group_order: # ditto
					die(4, 'Private key == secp256k1_group_order!')
				else: # return key mod group order as the key
					if not self.cfg.test_suite:
						ymsg(f'Warning: private key is greater than secp256k1 group order!:\n  {sec.hex()}')
					return (pk % self.secp256k1_group_order).to_bytes(self.privkey_len, 'big')

	class DummyWIF:
		"""
		Ethereum and Monero protocols inherit from this class
		"""
		def encode_wif(self, privbytes, pubkey_type, compressed):
			assert pubkey_type == self.pubkey_type, f'{pubkey_type}: invalid pubkey_type for {self.name} protocol!'
			assert compressed is False, f'{self.name} protocol does not support compressed pubkeys!'
			return privbytes.hex()

		def decode_wif(self, wif):
			return decoded_wif(
				sec         = bytes.fromhex(wif),
				pubkey_type = self.pubkey_type,
				compressed  = False)

def init_proto(
		cfg,
		coin       = None,
		*,
		testnet    = False,
		regtest    = False,
		network    = None,
		network_id = None,
		tokensym   = None,
		need_amt   = False,
		return_cls = False):

	assert type(testnet) is bool, 'init_proto_chk1'
	assert type(regtest) is bool, 'init_proto_chk2'
	assert coin or network_id, 'init_proto_chk3'
	assert not (coin and network_id), 'init_proto_chk4'

	if network_id:
		coin, network = CoinProtocol.Base.parse_network_id(network_id)
	elif network:
		assert network in CoinProtocol.Base.networks, f'init_proto_chk5 - {network!r}: invalid network'
		assert testnet is False, 'init_proto_chk6'
		assert regtest is False, 'init_proto_chk7'
	else:
		network = 'regtest' if regtest else 'testnet' if testnet else 'mainnet'

	coin = coin.lower()

	if coin not in CoinProtocol.coins:
		from .altcoin.params import init_genonly_altcoins
		init_genonly_altcoins(coin, testnet=testnet) # raises exception on failure

	name = CoinProtocol.coins[coin].name
	proto_name = name + ('' if network == 'mainnet' else network.capitalize())

	if not hasattr(CoinProtocol, proto_name):
		import importlib
		setattr(
			CoinProtocol,
			proto_name,
			getattr(importlib.import_module(f'mmgen.proto.{coin}.params'), network)
		)

	if return_cls:
		return getattr(CoinProtocol, proto_name)

	return getattr(CoinProtocol, proto_name)(
		cfg       = cfg,
		coin      = coin,
		name      = name,
		network   = network,
		tokensym  = tokensym,
		need_amt  = need_amt)

def init_proto_from_cfg(cfg, need_amt):
	return init_proto(
		cfg       = cfg,
		coin      = cfg.coin,
		network   = cfg.network,
		tokensym  = cfg.token,
		need_amt  = need_amt)

def warn_trustlevel(cfg):

	coinsym = cfg.coin

	if coinsym.lower() in CoinProtocol.coins:
		trust_level = CoinProtocol.coins[coinsym.lower()].trust_level
	else:
		from .altcoin.params import CoinInfo
		e = CoinInfo.get_entry(coinsym, 'mainnet')
		trust_level = e.trust_level if e else None
		if trust_level in (None, -1):
			from .util import die
			die(1, f'Coin {coinsym} is not supported by {gc.proj_name}')

	if trust_level > 3:
		return

	m = """
		Support for coin {c!r} is EXPERIMENTAL.  {a}
		assumes no responsibility for any loss of funds you may incur.
		This coin’s testing status: {t}
		Are you sure you want to continue?
	"""

	from .util import fmt
	from .color import red, yellow, green

	warning = fmt(m).strip().format(
		c = coinsym.upper(),
		t = {
			0: red('COMPLETELY UNTESTED'),
			1: red('LOW'),
			2: yellow('MEDIUM'),
			3: green('OK'),
		}[trust_level],
		a = gc.author)

	if cfg.test_suite:
		cfg._util.qmsg(warning)
		return

	from .ui import keypress_confirm
	if not keypress_confirm(cfg, warning, default_yes=True):
		import sys
		sys.exit(0)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.label: Monero wallet ops for the MMGen Suite
"""

from ...color import pink, cyan, gray
from ...util import msg, ymsg, gmsg, die, make_timestr
from ...ui import keypress_confirm
from ...addr import CoinAddr

from ..rpc import MoneroWalletRPC

from .spec import OpMixinSpec
from .wallet import OpWallet

class OpLabel(OpMixinSpec, OpWallet):
	spec_id  = 'label_spec'
	spec_key = ((1, 'source'),)
	opts     = ()
	wallet_offline = True

	async def main(self):

		gmsg('\n{a} label for wallet {b}, account #{c}, address #{d}'.format(
			a = 'Setting' if self.label else 'Removing',
			b = self.source.idx,
			c = self.account,
			d = self.address_idx
		))
		h = MoneroWalletRPC(self, self.source)

		h.open_wallet('source')
		wallet_data = h.get_wallet_data()

		max_acct = len(wallet_data.accts_data['subaddress_accounts']) - 1
		if self.account > max_acct:
			die(2, f'{self.account}: requested account index out of bounds (>{max_acct})')

		ret = h.print_acct_addrs(wallet_data, self.account)

		if self.address_idx > len(ret) - 1:
			die(2, '{}: requested address index out of bounds (>{})'.format(
				self.address_idx,
				len(ret) - 1))

		addr = ret[self.address_idx]
		new_label = f'{self.label} [{make_timestr()}]' if self.label else ''

		ca = CoinAddr(self.proto, addr['address'])
		from . import addr_width
		msg('\n  {a} {b}\n  {c} {d}\n  {e} {f}'.format(
				a = 'Address:       ',
				b = ca.hl(0) if self.cfg.full_address else ca.fmt(0, addr_width, color=True),
				c = 'Existing label:',
				d = pink(addr['label']) if addr['label'] else gray('[none]'),
				e = 'New label:     ',
				f = pink(new_label) if new_label else gray('[none]')))

		op = 'remove' if not new_label else 'update' if addr['label'] else 'set'

		if addr['label'] == new_label:
			ymsg('\nLabel is unchanged, operation cancelled')
		elif keypress_confirm(self.cfg, f'  {op.capitalize()} label?'):
			h.set_label(self.account, self.address_idx, new_label)
			ret = h.print_acct_addrs(h.get_wallet_data(print=False), self.account)
			label_chk = ret[self.address_idx]['label']
			if label_chk != new_label:
				ymsg(f'Warning: new label {label_chk!r} does not match requested value!')
				return False
			else:
				msg(cyan('\nLabel successfully {}'.format('set' if op == 'set' else op+'d')))
		else:
			ymsg('\nOperation cancelled by user request')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.dump: Monero wallet ops for the MMGen Suite
"""

from ...util import msg

from ..file.outputs import MoneroWalletDumpFile
from ..rpc import MoneroWalletRPC

from .wallet import OpWallet

class OpDump(OpWallet):
	wallet_offline = True

	async def process_wallet(self, d, fn, last):
		h = MoneroWalletRPC(self, d)
		h.open_wallet('source')
		wallet_data = h.get_wallet_data(print=False)
		msg('')
		MoneroWalletDumpFile.New(
			parent    = self,
			wallet_fn = fn,
			data      = {'wallet_metadata': wallet_data.addrs_data}
		).write()
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.base: THORChain base transaction class
"""

from ....tx.base import Base as TxBase

class Base(TxBase):

	dfl_fee = 2000000
	dfl_gas = 800000
	usr_contract_data = None
	disable_fee_check = False

	@property
	def nondata_outputs(self):
		return self.outputs

	@property
	def usr_fee(self):
		return self.proto.coin_amt(self.dfl_fee, from_unit='satoshi')

	def add_blockcount(self):
		pass

	def is_replaceable(self):
		return False

	def check_serialized_integrity(self):
		if self.signed:
			from .protobuf import RuneTx
			tx = RuneTx.loads(bytes.fromhex(self.serialized))

			o = self.txobj
			b = tx.body.messages[0].body

			if self.is_swap:
				from_k, amt_k = ('signer', 'coins')
				if b.memo != self.swap_memo:
					raise ValueError(f'{b.memo}: invalid swap memo in serialized data')
			else:
				from_k, amt_k = ('fromAddress', 'amount')
				if s := self.proto.encode_addr_bech32x(b.toAddress) != o['to']:
					raise ValueError(f'{s}: invalid ‘to’ address in serialized data')

			if s := self.proto.encode_addr_bech32x(getattr(b, from_k)) != o['from']:
				raise ValueError(f'{s}: invalid {from_k} in serialized data')

			if d := self.proto.coin_amt(int(getattr(b, amt_k)[0].amount), from_unit='satoshi') != o['amt']:
				raise ValueError(f'{d}: invalid send amount in serialized data')

			if n := tx.authInfo.signerInfos[0].sequence != o['sequence']:
				raise ValueError(f'{n}: invalid sequence number in serialized data')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.runeswap: THORChain swap tests for the cmdtest.py test suite
"""

from hashlib import md5

from mmgen.fileutil import get_data_from_file

from .httpd.thornode.swap import ThornodeSwapServer
from .include.proxy import TestProxy

from .regtest import CmdTestRegtest
from .swap import CmdTestSwapMethods, create_cross_methods
from .rune import CmdTestRune

class CmdTestRuneSwap(CmdTestSwapMethods, CmdTestRegtest):
	'RUNE swap operations'

	bdb_wallet = True
	tmpdir_nums = [57]
	networks = ('btc',)
	passthru_opts = ('coin', 'rpc_backend')
	cross_group = 'runeswap_rune'
	cross_coin = 'rune'

	cmd_group_in = (
		('setup',                'regtest (Bob and Alice) mode setup'),
		('subgroup.init',        []),
		('subgroup.rune_init',   ['init']),
		('subgroup.rune_swap',   ['rune_init']),
		('rune_rpc_server_stop', 'stopping the Thornode RPC server'),
		('swap_server_stop',     'stopping the Thornode swap server'),
		('stop',                 'stopping the regtest daemon'),
	)
	cmd_subgroups = {
		'init': (
			'creating Bob’s MMGen wallet and tracking wallet',
			('walletconv_bob', 'wallet creation (Bob)'),
			('addrgen_bob',    'address generation (Bob)'),
			('addrimport_bob', 'importing Bob’s addresses'),
		),
		'rune_init': (
			'initializing the RUNE tracking wallet',
			('rune_addrgen',     ''),
			('rune_addrimport',  ''),
			('rune_bal_refresh', ''),
			('rune_twview',      ''),
		),
		'rune_swap': (
			'swap operations (RUNE -> BTC)',
			('rune_swaptxcreate1',  ''),
			('rune_swaptxsign1',    ''),
			('rune_swaptxsend1',    ''),
			('rune_swaptxstatus1',  ''),
			('rune_swaptxreceipt1', ''),
			('rune_swaptxhex1',     ''),
		),
	}

	exec(create_cross_methods(cross_coin, cross_group, cmd_group_in, cmd_subgroups))

	def __init__(self, cfg, trunner, cfgs, spawn):

		super().__init__(cfg, trunner, cfgs, spawn)

		if not trunner:
			return

		globals()[self.cross_group] = self.create_cross_runner(trunner)

		self.swap_server = ThornodeSwapServer()
		self.swap_server.start()

		TestProxy(self, cfg)

	def swap_server_stop(self):
		return self._thornode_server_stop()

class CmdTestRuneSwapRune(CmdTestSwapMethods, CmdTestRune):
	'RUNE swap operations - RUNE wallet'

	networks = ('rune',)
	tmpdir_nums = [58]
	input_sels_prompt = 'to spend from: '
	is_helper = True
	txhex_chksum = '34980b41'

	cmd_group_in = CmdTestRune.cmd_group_in + (
		# rune_swap:
		('swaptxcreate1',            'creating a RUNE->BTC swap transaction'),
		('swaptxsign1',              'signing the transaction'),
		('swaptxsend1',              'sending the transaction'),
		('swaptxstatus1',            'getting the transaction status'),
		('swaptxreceipt1',           'getting the transaction receipt'),
		('swaptxhex1',               'dumping the transaction hex'),
		('thornode_server_stop',     'stopping Thornode server'),
	)

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.txhex_file = f'{self.tmpdir}/tx_dump.hex'

	def swaptxcreate1(self):
		t = self._swaptxcreate(['RUNE', '8.765', 'BTC'])
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t, inputs='3')

	def swaptxsign1(self):
		return self._swaptxsign()

	def swaptxsend1(self):
		return self._swaptxsend(add_opts=[f'--proxy=localhost:{TestProxy.port}'])

	def swaptxstatus1(self):
		return self._swaptxsend(add_opts=['--verbose', '--status'], status=True)

	def swaptxreceipt1(self):
		return self._swaptxsend(add_opts=['--receipt'], spawn_only=True)

	def swaptxhex1(self):
		t = self._swaptxsend(add_opts=[f'--dump-hex={self.txhex_file}'], dump_hex=True)
		t.read()
		txhex = get_data_from_file(self.cfg, self.txhex_file, silent=True)
		assert md5(txhex.encode()).hexdigest()[:8] == self.txhex_chksum
		return t



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
led: Control the LED on a single-board computer
"""

import sys, os, threading
from collections import namedtuple
from subprocess import run

from .util import msg, msg_r, die, have_sudo
from .color import blue, orange
from .base_obj import Lockable

class LEDControl:

	class binfo(Lockable):
		_reset_ok = ('trigger_reset',)

		def __init__(
				self,
				*,
				name,
				control,
				trigger         = None,
				trigger_dfl     = 'heartbeat',
				trigger_disable = 'none',
				color           = 'colored'):
			self.name = name
			self.control = control
			self.trigger = trigger
			self.trigger_dfl = trigger_dfl
			self.trigger_reset = trigger_dfl
			self.trigger_disable = trigger_disable
			self.color = color

	boards = {
		'raspi_pi': binfo(
			name    = 'Raspberry Pi',
			control = '/sys/class/leds/led0/brightness',
			trigger = '/sys/class/leds/led0/trigger',
			trigger_dfl = 'mmc0',
			color   = 'red'),
		'orange_pi': binfo(
			name    = 'Orange Pi (Armbian)',
			control = '/sys/class/leds/orangepi:red:status/brightness',
			color   = 'red'),
		'orange_pi_5': binfo(
			name    = 'Orange Pi 5 (Armbian)',
			control = '/sys/class/leds/status_led/brightness',
			color   = 'red'),
		'rock_pi': binfo(
			name    = 'Rock Pi (Armbian)',
			control = '/sys/class/leds/status/brightness',
			trigger = '/sys/class/leds/status/trigger',
			color   = 'blue'),
		'rock_5': binfo(
			name    = 'Rock 5 (Armbian)',
			control = '/sys/class/leds/user-led2/brightness',
			trigger = '/sys/class/leds/user-led2/trigger',
			color   = 'blue'),
		'banana_pi_f3': binfo(
			name    = 'Banana Pi F3 (Armbian)',
			control = '/sys/class/leds/sys-led/brightness',
			trigger = '/sys/class/leds/sys-led/trigger',
			color   = 'green'),
		'nano_pi_m6': binfo(
			name    = 'Nano Pi M6 (Armbian)',
			control = '/sys/class/leds/user_led/brightness',
			trigger = '/sys/class/leds/user_led/trigger',
			color   = 'green'),
		'dummy': binfo(
			name    = 'Fake Board',
			control = '/tmp/led_status',
			trigger = '/tmp/led_trigger'),
	}

	def __init__(self, *, enabled, simulate=False, debug=False):

		self.enabled = enabled
		self.debug = debug or simulate

		if not enabled:
			self.set = self.stop = self.noop
			return

		self.ev = threading.Event()
		self.led_thread = None

		for board_id, board in self.boards.items():
			if board_id == 'dummy' and not simulate:
				continue
			try:
				os.stat(board.control)
				break
			except FileNotFoundError:
				pass
		else:
			die('NoLEDSupport', 'Control files not found!  LED control not supported on this system')

		msg(f'{board.name} board detected')

		if self.debug:
			msg(f'\n  Status file:  {board.control}\n  Trigger file: {board.trigger}')

		def write_init_val(fn, init_val):
			if not init_val:
				with open(fn) as fp:
					init_val = fp.read().strip()
			with open(fn, 'w') as fp:
				fp.write(f'{init_val}\n')

		def permission_error_action(fn, desc):
			cmd = f'sudo chmod 0666 {fn}'
			if have_sudo():
				msg(orange(f'Running ‘{cmd}’'))
				run(cmd.split(), check=True)
			else:
				msg('\n{}\n{}\n{}'.format(
					blue(f'You do not have write access to the {desc}'),
					blue(f'To allow access, run the following command:\n\n    {cmd}'),
					orange('[To prevent this message in the future, enable sudo without a password]')
				))
				sys.exit(1)

		def init_state(fn, *, desc, init_val=None):
			try:
				write_init_val(fn, init_val)
			except PermissionError:
				permission_error_action(fn, desc)
				write_init_val(fn, init_val)

		# Writing to control file can alter trigger file, so read and initialize trigger file first:
		if board.trigger:
			def get_cur_state():
				try:
					with open(board.trigger) as fh:
						states = fh.read()
				except PermissionError:
					permission_error_action(board.trigger, 'status LED trigger file')
					with open(board.trigger) as fh:
						states = fh.read()

				res = [a for a in states.split() if a.startswith('[') and a.endswith(']')]
				return res[0][1:-1] if len(res) == 1 else None

			if cur_state := get_cur_state():
				msg(f'Saving current LED trigger state: [{cur_state}]')
				board.trigger_reset = cur_state
			else:
				msg('Unable to determine current LED trigger state')

			init_state(board.trigger, desc='status LED trigger file', init_val=board.trigger_disable)

		init_state(board.control, desc='status LED control file')

		self.board = board

	@classmethod
	def create_dummy_control_files(cls):
		db = cls.boards['dummy']
		with open(db.control, 'w') as fp:
			fp.write('0\n')
		with open(db.trigger, 'w') as fp:
			fp.write(db.trigger_dfl + '\n')

	def noop(self, *args, **kwargs):
		pass

	def ev_sleep(self, secs):
		self.ev.wait(secs)
		return self.ev.is_set()

	def led_loop(self, on_secs, off_secs):

		if self.debug:
			msg(f'led_loop({on_secs}, {off_secs})')

		if not on_secs:
			with open(self.board.control, 'w') as fp:
				fp.write('0\n')
			while True:
				if self.ev_sleep(3600):
					return

		while True:
			for s_time, val in ((on_secs, 255), (off_secs, 0)):
				if self.debug:
					msg_r(('^', '+')[bool(val)])
				with open(self.board.control, 'w') as fp:
					fp.write(f'{val}\n')
				if self.ev_sleep(s_time):
					if self.debug:
						msg('\n')
					return

	def set(self, state):
		lt = namedtuple('led_timings', ['on_secs', 'off_secs'])
		timings = {
			'off':     lt(0,    0),
			'standby': lt(2.2,  0.2),
			'busy':    lt(0.06, 0.06),
			'error':   lt(0.5,  0.5)}

		if self.led_thread:
			self.ev.set()
			self.led_thread.join()
			self.ev.clear()

		if self.debug:
			msg(f'Setting LED state to {state!r}')

		self.led_thread = threading.Thread(
				target = self.led_loop,
				name   = 'LED loop',
				args   = timings[state],
				daemon = True)

		self.led_thread.start()

	def stop(self):

		self.set('off')
		self.ev.set()
		self.led_thread.join()

		if self.debug:
			msg('Stopping LED')

		if self.board.trigger:
			with open(self.board.trigger, 'w') as fp:
				fp.write(self.board.trigger_reset + '\n')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.unsigned: Ethereum unsigned transaction class
"""

import json

from ....tx import unsigned as TxBase
from ....obj import CoinTxID, ETHNonce, Int, HexStr
from ....addr import CoinAddr, ContractAddr

from ...vm.tx.unsigned import Unsigned as VmUnsigned

from ..contract import Token, THORChainRouterContract

from .completed import Completed, TokenCompleted

class Unsigned(VmUnsigned, Completed, TxBase.Unsigned):

	def parse_txfile_serialized_data(self):
		d = self.serialized if isinstance(self.serialized, dict) else json.loads(self.serialized)
		o = {
			'from':     CoinAddr(self.proto, d['from']),
			# NB: for token, 'to' is sendto address
			'to':       CoinAddr(self.proto, d['to']) if d['to'] else None,
			'amt':      self.proto.coin_amt(d['amt']),
			'gasPrice': self.proto.coin_amt(d['gasPrice']),
			'startGas': (
				self.proto.coin_amt(d['startGas']).toWei() if '.' in d['startGas'] # for backward compat
				else int(d['startGas'])),
			'nonce':    ETHNonce(d['nonce']),
			'chainId':  None if d['chainId'] == 'None' else Int(d['chainId']),
			'data':     HexStr(d['data'])}
		self.txobj = o
		return d # 'token_addr', 'decimals' required by Token subclass

	async def do_sign(self, o, wif):
		o_conv = {
			'to':       bytes.fromhex(o['to'] or ''),
			'startgas': o['startGas'],
			'gasprice': o['gasPrice'].toWei(),
			'value':    o['amt'].toWei() if o['amt'] else 0,
			'nonce':    o['nonce'],
			'data':     self.swap_memo.encode() if self.is_swap else bytes.fromhex(o['data'])}

		from .transaction import Transaction
		etx = Transaction(**o_conv).sign(wif, o['chainId'])
		assert etx.sender.hex() == o['from'], (
			'Sender address recovered from signature does not match true sender')

		from .. import rlp
		self.serialized = rlp.encode(etx).hex()
		self.coin_txid = CoinTxID(etx.hash.hex())

		if o['data']: # contract-creating transaction
			if o['to']:
				if not self.is_swap:
					raise ValueError('contract-creating transaction cannot have to-address')
			else:
				self.txobj['token_addr'] = ContractAddr(self.proto, etx.creates.hex())

class TokenUnsigned(TokenCompleted, Unsigned):
	desc = 'unsigned transaction'

	def parse_txfile_serialized_data(self):
		d = Unsigned.parse_txfile_serialized_data(self)
		o = self.txobj
		o['token_addr'] = ContractAddr(self.proto, d['token_addr'])
		o['decimals'] = Int(d['decimals'])
		o['token_to'] = o['to']
		if self.is_swap:
			o['expiry'] = Int(d['expiry'])
			o['router_gas'] = Int(d['router_gas'])

	async def do_sign(self, o, wif):
		t = Token(self.cfg, self.proto, o['token_addr'], decimals=o['decimals'])
		tdata = t.create_transfer_data(o['to'], o['amt'], op=self.token_op)
		tx_in = t.make_tx_in(gas=o['startGas'], gasPrice=o['gasPrice'], nonce=o['nonce'], data=tdata)
		res = await t.txsign(tx_in, wif, o['from'], chain_id=o['chainId'])
		self.serialized = res.txhex
		self.coin_txid = res.txid
		if self.is_swap:
			c = THORChainRouterContract(self.cfg, self.proto, o['to'], decimals=o['decimals'])
			cdata = c.create_deposit_with_expiry_data(
				self.token_vault_addr,
				o['token_addr'],
				o['amt'],
				self.swap_memo.encode(),
				o['expiry'])
			tx_in = c.make_tx_in(
				gas = o['router_gas'],
				gasPrice = o['gasPrice'],
				nonce = o['nonce'] + 1,
				data = cdata)
			res = await t.txsign(tx_in, wif, o['from'], chain_id=o['chainId'])
			self.serialized2 = res.txhex
			self.coin_txid2 = res.txid

class AutomountUnsigned(TxBase.AutomountUnsigned, Unsigned):
	pass

class TokenAutomountUnsigned(TxBase.AutomountUnsigned, TokenUnsigned):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.secp256k1.util: secp256k1 elliptic curve utility functions
"""

def sign_message(*, sign_doc, sec_bytes):
	from hashlib import sha256
	from .secp256k1 import sign_msghash
	return sign_msghash(
		sha256(bytes(sign_doc)).digest(),
		sec_bytes)[0]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/objattrtest.py: Test immutable attributes of MMGen data objects
"""

# TODO: test 'typeconv' during instance creation

from collections import namedtuple

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from mmgen.cfg import Config
from mmgen.util import msg, msg_r, gmsg, die
from mmgen.color import red, yellow, green, blue, purple, nocolor
from mmgen.obj import ListItemAttr

opts_data = {
	'sets': [
		('show_descriptor_type', True, 'verbose', True),
	],
	'text': {
		'desc': 'Test immutable attributes of MMGen data objects',
		'usage':'[options] [object]',
		'options': """
-h, --help                  Print this help message
--, --longhelp              Print help message for long (global) options
-d, --show-descriptor-type  Display the attribute's descriptor type
-v, --verbose               Produce more verbose output
"""
	}
}

cfg = Config(opts_data=opts_data)

from test.include.common import set_globals
set_globals(cfg)

from test.objattrtest_d.common import sample_objs

pd = namedtuple('attr_bits', ['read_ok', 'delete_ok', 'reassign_ok', 'typeconv', 'set_none_ok'])
perm_bits = ('read_ok', 'delete_ok', 'reassign_ok')
attr_dfls = {
	'reassign_ok': False,
	'delete_ok': False,
	'typeconv': True,
	'set_none_ok': False,
}

def parse_attrbits(bits):
	return pd(
		bool(0b00001 & bits), # read
		bool(0b00010 & bits), # delete
		bool(0b00100 & bits), # reassign
		bool(0b01000 & bits), # typeconv
		bool(0b10000 & bits), # set_none
	)

def get_descriptor_obj(objclass, attrname):
	for o in (objclass, objclass.__bases__[0]): # assume there's only one base class
		if attrname in o.__dict__:
			return o.__dict__[attrname]
	die(4, f'unable to find descriptor {objclass.__name__}.{attrname}')

def test_attr_perm(obj, attrname, perm_name, perm_value, dobj, attrval_type):

	class SampleObjError(Exception):
		pass

	pname = perm_name.replace('_ok', '')
	pstem = pname.rstrip('e')

	try:
		if perm_name == 'read_ok': # non-existent perm
			getattr(obj, attrname)
		elif perm_name == 'reassign_ok':
			try:
				so = sample_objs[attrval_type.__name__]
			except Exception as e:
				raise SampleObjError(f'unable to find sample object of type {attrval_type.__name__!r}') from e
			# ListItemAttr allows setting an attribute if its value is None
			if type(dobj) is ListItemAttr and getattr(obj, attrname) is None:
				setattr(obj, attrname, so)
			setattr(obj, attrname, so)
		elif perm_name == 'delete_ok':
			delattr(obj, attrname)
	except SampleObjError as e:
		die(4, f'Test script error ({e})')
	except Exception as e:
		if perm_value is True:
			fs = '{!r}: unable to {} attribute {!r}, though {}ing is allowed ({})'
			die(4, fs.format(type(obj).__name__, pname, attrname, pstem, e))
	else:
		if perm_value is False:
			fs = '{!r}: attribute {!r} is {n}able, though {n}ing is forbidden'
			die(4, fs.format(type(obj).__name__, attrname, n=pstem))

def test_attr(data, obj, attrname, dobj, bits, attrval_type):
	if hasattr(obj, attrname): # TODO
		td_attrval_type = data.attrs[attrname][1]

		if attrval_type not in (td_attrval_type, type(None)):
			fs = '\nattribute {!r} of {!r} instance has incorrect type {!r} (should be {!r})'
			die(4, fs.format(attrname, type(obj).__name__, attrval_type.__name__, td_attrval_type.__name__))

	if hasattr(dobj, '__dict__'):
		d = dobj.__dict__
		bits = bits._asdict()
		colors = {
			'reassign_ok': purple,
			'delete_ok': red,
			'typeconv': green,
			'set_none_ok': yellow,
		}
		for k in bits:
			if k in d:
				if d[k] != bits[k]:
					fs = 'init value {iv}={a} for attr {n!r} does not match test data ({iv}={b})'
					die(4, fs.format(iv=k, n=attrname, a=d[k], b=bits[k]))
				if cfg.verbose and d[k] != attr_dfls[k]:
					msg_r(colors[k](f' {k}={d[k]!r}'))

def test_object(mod, test_data, objname):

	if '.' in objname:
		on1, on2 = objname.split('.')
		cls = getattr(getattr(mod, on1), on2)
	else:
		cls = getattr(mod, objname)

	fs = 'Testing attribute ' + ('{!r:<15}{dt:13}' if cfg.show_descriptor_type else '{!r}')
	data = test_data[objname]
	obj = cls(*data.args, **data.kwargs)

	for attrname, adata in data.attrs.items():
		dobj = get_descriptor_obj(type(obj), attrname)
		if cfg.verbose:
			msg_r(fs.format(attrname, dt=type(dobj).__name__.replace('MMGen', '')))
		bits = parse_attrbits(adata[0])
		test_attr(data, obj, attrname, dobj, bits, adata[1])
		for bit_name, bit_value in bits._asdict().items():
			if bit_name in perm_bits:
				test_attr_perm(obj, attrname, bit_name, bit_value, dobj, adata[1])
		cfg._util.vmsg('')

def do_loop():
	import importlib
	modname = f'test.objattrtest_d.{proto.coin.lower()}_{proto.network}'
	mod = importlib.import_module(modname)
	test_data = getattr(mod, 'tests')
	gmsg(f'Running immutable attribute tests for {proto.coin} {proto.network}')

	utests = cfg._args
	for obj in test_data:
		if utests and obj not in utests:
			continue
		msg((blue if cfg.verbose else nocolor)(f'Testing {obj}'))
		test_object(mod, test_data, obj)

proto = cfg._proto

if __name__ == '__main__':
	do_loop()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.zec.params: Zcash protocol
"""

from ..btc.params import mainnet
from ...protocol import decoded_addr
from ...addr import CoinAddr

class ZcashViewKey(CoinAddr):
	hex_width = 128

class mainnet(mainnet):
	base_coin      = 'ZEC'
	addr_ver_info  = {'1cb8': 'p2pkh', '1cbd': 'p2sh', '169a': 'zcash_z', 'a8abd3': 'viewkey'}
	wif_ver_num    = {'std': '80', 'zcash_z': 'ab36'}
	pubkey_types   = ('std', 'zcash_z')
	mmtypes        = ('L', 'C', 'Z')
	mmcaps         = ()
	dfl_mmtype     = 'L'
	avg_bdi        = 75

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.coin_id = 'ZEC-Z' if self.cfg.type in ('zcash_z', 'Z') else 'ZEC-T'

	def get_wif_ver_bytes_len(self, key_data):
		"""
		vlen must be set dynamically since Zcash has variable-length version bytes
		"""
		for v in self.wif_ver_bytes.values():
			if key_data[:len(v)] == v:
				return len(v)
		raise ValueError('Invalid WIF version number')

	def get_addr_len(self, addr_fmt):
		return (20, 64)[addr_fmt in ('zcash_z', 'viewkey')]

	def decode_addr_bytes(self, addr_bytes):
		"""
		vlen must be set dynamically since Zcash has variable-length version bytes
		"""
		for ver_bytes, addr_fmt in self.addr_ver_bytes.items():
			vlen = len(ver_bytes)
			if addr_bytes[:vlen] == ver_bytes:
				if len(addr_bytes[vlen:]) == self.get_addr_len(addr_fmt):
					return decoded_addr(addr_bytes[vlen:], ver_bytes, addr_fmt)

		return False

	def preprocess_key(self, sec, pubkey_type):
		if pubkey_type == 'zcash_z': # zero the first four bits
			return bytes([sec[0] & 0x0f]) + sec[1:]
		else:
			return super().preprocess_key(sec, pubkey_type)

	def pubhash2addr(self, pubhash, addr_type):
		hash_len = len(pubhash)
		if hash_len == 20:
			return super().pubhash2addr(pubhash, addr_type)
		elif hash_len == 64:
			raise NotImplementedError('Zcash z-addresses do not support pubhash2addr()')
		else:
			raise ValueError(f'{hash_len}: incorrect pubkey hash length')

	def viewkey(self, viewkey_str):
		return ZcashViewKey.__new__(ZcashViewKey, self, viewkey_str)

class testnet(mainnet):
	wif_ver_num  = {'std': 'ef', 'zcash_z': 'ac08'}
	addr_ver_info = {'1d25': 'p2pkh', '1cba': 'p2sh', '16b6': 'zcash_z', 'a8ac0c': 'viewkey'}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.restore: Monero wallet ops for the MMGen Suite
"""

from ...util import msg, gmsg, bmsg, ymsg, rmsg, die

from ..file.outputs import MoneroWalletDumpFile
from ..rpc import MoneroWalletRPC

from .create import OpCreate

class OpRestore(OpCreate):
	wallet_offline = True

	def check_uopts(self):
		if self.cfg.restore_height is not None:
			die(1, '--restore-height must be unset when running the ‘restore’ command')

	async def process_wallet(self, d, fn, last):

		def get_dump_data():
			def gen():
				for fn in [self.get_wallet_fn(d, watch_only=wo) for wo in (True, False)]:
					ret = fn.parent / (fn.name + '.dump')
					if ret.exists():
						yield ret
			dump_fns = tuple(gen())
			if not dump_fns:
				die(1, f"No suitable dump file found for '{fn}'")
			elif len(dump_fns) > 1:
				ymsg(f"Warning: more than one dump file found for '{fn}' - using the first!")
			return MoneroWalletDumpFile.Completed(
				parent = self,
				fn     = dump_fns[0]).data._asdict()['wallet_metadata']

		def restore_accounts():
			bmsg('  Restoring accounts:')
			for acct_idx, acct_data in enumerate(data[1:], 1):
				msg(fs.format(acct_idx, 0, acct_data['address']))
				self.c.call('create_account')

		def restore_subaddresses():
			bmsg('  Restoring subaddresses:')
			for acct_idx, acct_data in enumerate(data):
				for addr_idx, addr_data in enumerate(acct_data['addresses'][1:], 1):
					msg(fs.format(acct_idx, addr_idx, addr_data['address']))
					self.c.call('create_address', account_index=acct_idx)

		def restore_labels():
			bmsg('  Restoring labels:')
			for acct_idx, acct_data in enumerate(data):
				for addr_idx, addr_data in enumerate(acct_data['addresses']):
					addr_data['used'] = False # do this so that restored data matches
					msg(fs.format(acct_idx, addr_idx, addr_data['label']))
					self.c.call(
						'label_address',
						index = { 'major': acct_idx, 'minor': addr_idx },
						label = addr_data['label'],
					)

		def make_format_str():
			return '    acct {:O>%s}, addr {:O>%s} [{}]' % (
				len(str(len(data) - 1)),
				len(str(max(len(acct_data['addresses']) for acct_data in data) - 1))
			)

		def check_restored_data():
			restored_data = h.get_wallet_data(print=False).addrs_data
			if restored_data != data:
				rmsg('Restored data does not match original dump!  Dumping bad data.')
				MoneroWalletDumpFile.New(
					parent    = self,
					wallet_fn = fn,
					data      = {'wallet_metadata': restored_data}
				).write(add_suf='.bad')
				die(3, 'Fatal error')

		await super().process_wallet(d, fn, last)

		h = MoneroWalletRPC(self, d)
		h.open_wallet('newly created')

		msg('')
		data = get_dump_data()
		fs = make_format_str()

		gmsg('\nRestoring accounts, subaddresses and labels from dump file:\n')

		restore_accounts()
		restore_subaddresses()
		restore_labels()

		check_restored_data()

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/tooltest.py: Tests for the 'mmgen-tool' utility
"""

import sys, os, time
from subprocess import run, PIPE

try:
	from include.test_init import repo_root
except ImportError:
	from test.include.test_init import repo_root

from mmgen.cfg import Config
from mmgen.color import red, yellow, green, blue, cyan
from mmgen.util import msg, msg_r, Msg, die

opts_data = {
	'text': {
		'desc': "Test suite for the 'mmgen-tool' utility",
		'usage':'[options] [command]',
		'options': """
-h, --help          Print this help message
-C, --coverage      Produce code coverage info using trace module
-d, --debug         Produce debugging output (stderr from spawned script)
--, --longhelp      Print help message for long (global) options
-l, --list-cmds     List and describe the tests and commands in this test suite
-s, --testing-status  List the testing status of all 'mmgen-tool' commands
-t, --type=t        Specify address type (valid choices: 'zcash_z')
-v, --verbose       Produce more verbose output
""",
	'notes': """

If no command is given, the whole suite of tests is run.
"""
	}
}

sys.argv = [sys.argv[0]] + ['--skip-cfg-file'] + sys.argv[1:]

cfg = Config(opts_data=opts_data)

from test.include.common import (
	set_globals,
	mk_tmpdir,
	cleandir,
	write_to_tmpfile,
	ok,
	read_from_file,
	read_from_tmpfile,
	cmp_or_die,
	getrand,
	getrandhex,
	end_msg,
	init_coverage,
	get_tmpfile,
)
set_globals(cfg)

vmsg = cfg._util.vmsg

proto = cfg._proto

assert cfg.type in (None, 'zcash_z'), 'Only zcash-z permitted for --type argument'

cmd_data = {
	'cryptocoin': {
		'desc': 'Cryptocoin address/key commands',
		'cmd_data': {
			'randwif':        (),
			'randpair':       (), # create 4 pairs: uncomp, comp, segwit, bech32
			'wif2addr':       ('randpair', 'o4'),
			'wif2hex':        ('randpair', 'o4'),
			'privhex2pubhex': ('wif2hex', 'o3'),        # segwit only
			'pubhex2addr':    ('privhex2pubhex', 'o3'), # segwit only
			'hex2wif':        ('wif2hex', 'io2'),       # uncomp, comp
			'addr2pubhash':   ('randpair', 'o4'),       # uncomp, comp, bech32
			'pubhash2addr':   ('addr2pubhash', 'io4'),  # uncomp, comp, bech32
		},
	},
	'mnemonic': {
		'desc': 'mnemonic commands',
		'cmd_data': {
			'hex2mn':       (),
			'mn2hex':       ('hex2mn', 'io3'),
			'mn_rand128':   (),
			'mn_rand192':   (),
			'mn_rand256':   (),
			'mn_stats':     (),
			'mn_printlist': (),
		},
	},
}

if proto.coin in ('BTC', 'LTC'):
	cmd_data['cryptocoin']['cmd_data'].update({
		'pubhex2redeem_script': ('privhex2pubhex', 'o3'),
		'wif2redeem_script':    ('randpair', 'o3'),
		'wif2segwit_pair':      ('randpair', 'o2'),
		'privhex2addr':         ('wif2hex', 'o4'), # compare with output of randpair
		'pipetest':             ('randpair', 'o3')
	})

if proto.coin == 'XMR' or cfg.type == 'zcash_z':
	del cmd_data['cryptocoin']['cmd_data']['pubhash2addr']
	del cmd_data['cryptocoin']['cmd_data']['addr2pubhash']

tcfg = {
	'name':          'the tool utility',
	'enc_passwd':    'Ten Satoshis',
	'tmpdir':        'test/tmp/10',
	'tmpdir_num':    10,
	'refdir':        'test/ref',
	'txfile': {
		'btc': ('0B8D5A[15.31789,14,tl=1320969600].rawtx',
				'0C7115[15.86255,14,tl=1320969600].testnet.rawtx'),
		'bch': ('460D4D-BCH[10.19764,tl=1320969600].rawtx',
				'359FD5-BCH[6.68868,tl=1320969600].testnet.rawtx'),
		'ltc': ('AF3CDF-LTC[620.76194,1453,tl=1320969600].rawtx',
				'A5A1E0-LTC[1454.64322,1453,tl=1320969600].testnet.rawtx'),
	},
	'addrfile': '98831F3A{}[1,31-33,500-501,1010-1011]{}.addrs',
	'addrfile_chk':  {
		'btc': ('6FEF 6FB9 7B13 5D91','424E 4326 CFFE 5F51'),
		'bch': ('6FEF 6FB9 7B13 5D91','424E 4326 CFFE 5F51'),
		'ltc': ('AD52 C3FE 8924 AAF0','4EBE 2E85 E969 1B30'),
	}
}

ref_subdir  = '' if proto.base_coin == 'BTC' else proto.name.lower()
altcoin_pfx = '' if proto.base_coin == 'BTC' else '-'+proto.base_coin
tn_ext = ('', '.testnet')[proto.testnet]

spawn_cmd = [
	'scripts/exec_wrapper.py',
	os.path.relpath(os.path.join(repo_root, 'cmds', 'mmgen-tool'))]

if cfg.coverage:
	d, f = init_coverage()
	spawn_cmd = ['python3', '-m', 'trace', '--count', '--coverdir='+d, '--file='+f] + spawn_cmd
elif sys.platform == 'win32':
	spawn_cmd = ['python3'] + spawn_cmd

add_spawn_args = ['--data-dir='+tcfg['tmpdir']] + ['--{}{}'.format(
		k.replace('_', '-'),
		'='+getattr(cfg, k) if getattr(cfg, k) is not True else '')
			for k in ('testnet', 'rpc_host', 'regtest', 'coin', 'type') if getattr(cfg, k)]

if cfg.list_cmds:
	fs = '  {:<{w}} - {}'
	Msg('Available commands:')
	w = max(map(len, cmd_data))
	for cmd in cmd_data:
		Msg(fs.format(cmd, cmd_data[cmd]['desc'], w=w))
	Msg('\nAvailable utilities:')
	Msg(fs.format('clean', 'Clean the tmp directory', w=w))
	sys.exit(0)

if cfg.testing_status:
	tested_in = {
		'tooltest.py': [],
		'cmdtest.py': (
			'encrypt', 'decrypt', 'find_incog_data',
			'addrfile_chksum', 'keyaddrfile_chksum', 'passwdfile_chksum',
			'add_label', 'remove_label', 'remove_address', 'twview',
			'getbalance', 'listaddresses', 'listaddress',
			'daemon_version', 'decrypt_keystore', 'decrypt_geth_keystore',
			'mn2hex_interactive', 'rand2file',
			'rescan_address', 'rescan_blockchain', 'resolve_address',
			'twexport', 'twimport', 'txhist'
		),
		'tooltest2.py': run(
			['python3', 'test/tooltest2.py', '--list-tested-cmds'],
			stdout = PIPE,
			check = True
		).stdout.decode().split()
	}
	for v in cmd_data.values():
		tested_in['tooltest.py'] += list(v['cmd_data'].keys())

	Msg(green("Testing status of 'mmgen-tool' commands:"))
	for l in ('tooltest.py', 'tooltest2.py', 'cmdtest.py'):
		Msg('\n  ' + blue(l+':'))
		Msg('    '+'\n    '.join(sorted(tested_in[l])))

	ignore = ()
	from mmgen.main_tool import get_cmds
	uc = sorted(
		set(get_cmds()) -
		set(ignore) -
		set(tested_in['tooltest.py']) -
		set(tested_in['tooltest2.py']) -
		set(tested_in['cmdtest.py'])
	)
	if uc:
		Msg(yellow('\n  {}\n    {}'.format('Untested commands:', '\n    '.join(uc))))
	sys.exit(0)

from mmgen.key import is_wif
from mmgen.addr import is_coin_addr

def is_wif_loc(s):
	return is_wif(proto, s)

def is_coin_addr_loc(s):
	return is_coin_addr(proto, s)

msg_w = 35

def test_msg(m):
	msg_r(green(f'Testing {m}\n') if cfg.verbose else '{:{w}}'.format(f'Testing {m}', w=msg_w+8))

compressed = cfg.type or ('', 'compressed')['C' in proto.mmtypes]
segwit     = ('', 'segwit')['S' in proto.mmtypes]
bech32     = ('', 'bech32')['B' in proto.mmtypes]
type_compressed_arg = ([], ['--type=' + (cfg.type or 'compressed')])[bool(cfg.type) or 'C' in proto.mmtypes]
type_segwit_arg     = ([], ['--type=segwit'])['S' in proto.mmtypes]
type_bech32_arg     = ([], ['--type=bech32'])['B' in proto.mmtypes]

class MMGenToolTestUtils:

	def run_cmd(
			self,
			name,
			tool_args,
			kwargs    = '',
			extra_msg = '',
			silent    = False,
			strip     = True,
			add_opts  = [],
			binary    = False):
		sys_cmd = (
			spawn_cmd +
			add_spawn_args +
			['-r0', '-d', tcfg['tmpdir']] +
			add_opts +
			[name.lower()] +
			tool_args +
			kwargs.split()
		)
		if extra_msg:
			extra_msg = f'({extra_msg})'
		full_name = ' '.join([name.lower()]+add_opts+kwargs.split()+extra_msg.split())
		if not silent:
			if cfg.verbose:
				sys.stderr.write(green(f'Testing {full_name}\nExecuting '))
				sys.stderr.write(cyan(' '.join(sys_cmd)+'\n'))
			else:
				msg_r('Testing {:{w}}'.format(full_name+':', w=msg_w))

		cp = run(sys_cmd, stdout=PIPE, stderr=PIPE)
		out = cp.stdout
		err = cp.stderr
		if cfg.debug:
			from test.include.common import dmsg
			try:
				dmsg(err.decode())
			except:
				dmsg(repr(err))
		if not binary:
			out = out.decode()
		if cp.returncode != 0:
			msg('{}\n{}\n{}'.format(
				red('FAILED'),
				yellow('Command stderr output:'),
				err.decode()))
			die(2, f'Called process returned with an error (retcode {cp.returncode})')
		return (out, out.rstrip())[bool(strip)]

	def run_cmd_chk(self, name, f1, f2, kwargs='', extra_msg='', strip_hex=False, add_opts=[]):
		idata = read_from_file(f1).rstrip()
		odata = read_from_file(f2).rstrip()
		ret = self.run_cmd(name, [odata], kwargs=kwargs, extra_msg=extra_msg, add_opts=add_opts)
		vmsg('In:   ' + repr(odata))
		vmsg('Out:  ' + repr(ret))
		def cmp_equal(a, b):
			return (a.lstrip('0') == b.lstrip('0')) if strip_hex else (a == b)
		if cmp_equal(ret, idata):
			ok()
		else:
			die(4, f"Error: values don't match:\nIn:  {idata!r}\nOut: {ret!r}")
		return ret

	def run_cmd_nochk(self, name, f1, kwargs='', add_opts=[]):
		odata = read_from_file(f1).rstrip()
		ret = self.run_cmd(name, [odata], kwargs=kwargs, add_opts=add_opts)
		vmsg('In:   ' + repr(odata))
		vmsg('Out:  ' + repr(ret))
		return ret

	def run_cmd_out(
			self,
			name,
			carg      = None,
			Return    = False,
			kwargs    = '',
			fn_idx    = '',
			extra_msg = '',
			literal   = False,
			chkdata   = '',
			hush      = False,
			add_opts  = []):
		if carg:
			write_to_tmpfile(tcfg, f'{name}{fn_idx}.in', carg+'\n')
		ret = self.run_cmd(
				name,
				([], [carg])[bool(carg)],
				kwargs    = kwargs,
				extra_msg = extra_msg,
				add_opts  = add_opts)
		if carg:
			vmsg('In:   ' + repr(carg))
		vmsg('Out:  ' + (repr(ret), ret)[literal])
		if ret or ret == '':
			write_to_tmpfile(tcfg, f'{name}{fn_idx}.out', ret+'\n')
			if chkdata:
				cmp_or_die(ret, chkdata)
				return
			if Return:
				return ret
			elif not hush:
				ok()
		else:
			die(4, f'Error for command {name!r}')

	def run_cmd_randinput(self, name, strip=True, add_opts=[]):
		s = getrand(128)
		fn = name+'.in'
		write_to_tmpfile(tcfg, fn, s, binary=True)
		ret = self.run_cmd(name, [get_tmpfile(tcfg, fn)], strip=strip, add_opts=add_opts)
		fn = name+'.out'
		write_to_tmpfile(tcfg, fn, ret+'\n')
		ok()
		vmsg(f'Returned: {ret}')

tu = MMGenToolTestUtils()

def ok_or_die(val, chk_func, s, skip_ok=False):
	try:
		ret = chk_func(val)
	except:
		ret = False
	if ret:
		if not skip_ok:
			ok()
	else:
		die(4, f'Returned value {val!r} is not a {s}')

class MMGenToolTestCmds:

	# Cryptocoin
	def randwif(self, name):
		for n, k in enumerate(['', compressed]):
			ao = ['--type='+k] if k else []
			ret = tu.run_cmd_out(name, add_opts=ao, Return=True, fn_idx=n+1)
			ok_or_die(ret, is_wif_loc, 'WIF key')
	def randpair(self, name):
		for n, k in enumerate(['', compressed, segwit, bech32]):
			ao = ['--type='+k] if k else []
			wif, addr = tu.run_cmd_out(name, add_opts=ao, Return=True, fn_idx=n+1, literal=True).split()
			ok_or_die(wif, is_wif_loc, 'WIF key', skip_ok=True)
			ok_or_die(addr, is_coin_addr_loc, 'Coin address')
	def wif2addr(self, name, f1, f2, f3, f4):
		for n, f, k in (
				(1, f1, ''),
				(2, f2, compressed),
				(3, f3, segwit),
				(4, f4, bech32)):
			ao = ['--type='+k] if k else []
			wif = read_from_file(f).split()[0]
			tu.run_cmd_out(name, wif, add_opts=ao, fn_idx=n)
	def wif2hex(self, name, f1, f2, f3, f4):
		for n, f, m in (
				(1, f1, ''),
				(2, f2, compressed),
				(3, f3, '{} for {}'.format(compressed or 'uncompressed', segwit or 'p2pkh')),
				(4, f4, '{} for {}'.format(compressed or 'uncompressed', bech32 or 'p2pkh'))):
			wif = read_from_file(f).split()[0]
			tu.run_cmd_out(name, wif, fn_idx=n, extra_msg=m)
	def privhex2addr(self, name, f1, f2, f3, f4):
		keys = [read_from_file(f).rstrip() for f in (f1, f2, f3, f4)]
		for n, k in enumerate(('', compressed, segwit, bech32)):
			ao = ['--type='+k] if k else []
			ret = tu.run_cmd(name, [keys[n]], add_opts=ao).rstrip()
			iaddr = read_from_tmpfile(tcfg, f'randpair{n+1}.out').split()[-1]
			vmsg(f'Out: {ret}')
			cmp_or_die(iaddr, ret)
			ok()
	def hex2wif(self, name, f1, f2, f3, f4):
		for fi, fo, k in (
				(f1, f2, ''),
				(f3, f4, compressed)):
			ao = ['--type='+k] if k else []
			tu.run_cmd_chk(name, fi, fo, add_opts=ao)
	def addr2pubhash(self, name, f1, f2, f3, f4):
		for n, f, m, ao in (
				(1, f1, '', []),
				(2, f2, 'from {}'.format(compressed or 'uncompressed'), []),
				(4, f4, '', type_bech32_arg)):
			addr = read_from_file(f).split()[-1]
			tu.run_cmd_out(name, addr, fn_idx=n, add_opts=ao, extra_msg=m)
	def pubhash2addr(self, name, f1, f2, f3, f4, f5, f6, f7, f8):
		for _, fi, fo, m, ao in (
				(1, f1, f2, '', []),
				(2, f3, f4, 'from {}'.format(compressed or 'uncompressed'), []),
				(4, f7, f8, '', type_bech32_arg)):
			tu.run_cmd_chk(name, fi, fo, add_opts=ao, extra_msg=m)
	def privhex2pubhex(self, name, f1, f2, f3): # from Hex2wif
		addr = read_from_file(f3).strip()
		tu.run_cmd_out(name, addr, add_opts=type_compressed_arg, fn_idx=3) # what about uncompressed?
	def pubhex2redeem_script(self, name, f1, f2, f3): # from above
		addr = read_from_file(f3).strip()
		tu.run_cmd_out(name, addr, add_opts=type_segwit_arg, fn_idx=3)
		rs = read_from_tmpfile(tcfg, 'privhex2pubhex3.out').strip()
		tu.run_cmd_out('pubhex2addr', rs, add_opts=type_segwit_arg, fn_idx=3, hush=True)
		addr1 = read_from_tmpfile(tcfg, 'pubhex2addr3.out').strip()
		addr2 = read_from_tmpfile(tcfg, 'randpair3.out').split()[1]
		cmp_or_die(addr1, addr2)
		ok()
	def wif2redeem_script(self, name, f1, f2, f3): # compare output with above
		wif = read_from_file(f3).split()[0]
		ret1 = tu.run_cmd_out(name, wif, add_opts=type_segwit_arg, fn_idx=3, Return=True)
		ret2 = read_from_tmpfile(tcfg, 'pubhex2redeem_script3.out').strip()
		cmp_or_die(ret1, ret2)
		ok()
	def wif2segwit_pair(self, name, f1, f2): # does its own checking, so just run
		wif = read_from_file(f2).split()[0]
		tu.run_cmd_out(name, wif, add_opts=type_segwit_arg, fn_idx=2)
	def pubhex2addr(self, name, f1, f2, f3):
		addr = read_from_file(f3).strip()
		tu.run_cmd_out(name, addr, add_opts=type_segwit_arg, fn_idx=3)

	def pipetest(self, name, f1, f2, f3):
		wif = read_from_file(f3).split()[0]
		cmd = (
			'{c} {a} wif2hex {wif}' +
			' | {c} {a} --type=compressed privhex2pubhex -' +
			' | {c} {a} --type=segwit pubhex2redeem_script -' +
			' | {c} {a} --type=segwit redeem_script2addr -').format(
					c   = ' '.join(spawn_cmd),
					a   = ' '.join(add_spawn_args),
					wif = wif)
		test_msg('command piping')
		if cfg.verbose:
			sys.stderr.write(green('Executing ') + cyan(cmd) + '\n')
		res = run(cmd, stdout=PIPE, shell=True).stdout.decode().strip()
		addr = read_from_tmpfile(tcfg, 'wif2addr3.out').strip()
		cmp_or_die(addr, res)
		ok()

	# Mnemonic
	def hex2mn(self, name):
		for n, size, m in (
				(1, 16, '128-bit'),
				(2, 24, '192-bit'),
				(3, 32, '256-bit')):
			hexnum = getrandhex(size)
			tu.run_cmd_out(name, hexnum, fn_idx=n, extra_msg=m)
	def mn2hex(self, name, f1, f2, f3, f4, f5, f6):
		for f_i, f_o, m in ((f1, f2, '128-bit'), (f3, f4, '192-bit'), (f5, f6, '256-bit')):
			tu.run_cmd_chk(name, f_i, f_o, extra_msg=m, strip_hex=True)
	def mn_rand128(self, name):
		tu.run_cmd_out(name)
	def mn_rand192(self, name):
		tu.run_cmd_out(name)
	def mn_rand256(self, name):
		tu.run_cmd_out(name)
	def mn_stats(self, name):
		tu.run_cmd_out(name)
	def mn_printlist(self, name):
		tu.run_cmd(name, [])
		ok()

# main()
start_time = int(time.time())
mk_tmpdir(tcfg['tmpdir'])

def gen_deps_for_cmd(cdata):
	fns = []
	if cdata:
		name, code = cdata
		io, count = (code[:-1], int(code[-1])) if code[-1] in '0123456789' else (code, 1)
		for c in range(count):
			fns += ['{}{}{}'.format(
				name,
				(c+1 if count > 1 else ''),
				('.in' if ch == 'i' else '.out'),
			) for ch in io]
	return fns

def do_cmds(cmd_group):
	tc = MMGenToolTestCmds()
	gdata = cmd_data[cmd_group]['cmd_data']
	for cmd in gdata:
		fns = gen_deps_for_cmd(gdata[cmd])
		cmdline = [cmd] + [os.path.join(tcfg['tmpdir'], fn) for fn in fns]
		getattr(tc, cmd)(*cmdline)

def main():
	if cfg._args:
		if len(cfg._args) != 1:
			die(1, 'Only one command may be specified')
		cmd = cfg._args[0]
		if cmd in cmd_data:
			cleandir(tcfg['tmpdir'], do_msg=True)
			msg('Running tests for {}:'.format(cmd_data[cmd]['desc']))
			do_cmds(cmd)
		elif cmd == 'clean':
			cleandir(tcfg['tmpdir'], do_msg=True)
			sys.exit(0)
		else:
			die(1, f'{cmd!r}: unrecognized command')
	else:
		cleandir(tcfg['tmpdir'], do_msg=True)
		for cmd in cmd_data:
			msg('Running tests for {}:'.format(cmd_data[cmd]['desc']))
			do_cmds(cmd)
			if cmd is not list(cmd_data.keys())[-1]:
				msg('')
	end_msg(int(time.time()) - start_time)

from mmgen.main import launch
launch(func=main)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xmr.addrgen: Monero address generation class for the MMGen suite
"""

from ...addrgen import addr_generator, check_data
from ...addr import CoinAddr

class monero(addr_generator.keccak):

	def b58enc(self, addr_bytes):
		from ...baseconv import baseconv
		enc = baseconv('b58').frombytes
		l = len(addr_bytes)
		a = ''.join([enc(addr_bytes[i*8:i*8+8], pad=11, tostr=True) for i in range(l//8)])
		b = enc(addr_bytes[l-l%8:], pad=7, tostr=True)
		return a + b

	@check_data
	def to_addr(self, data):
		step1 = self.proto.addr_fmt_to_ver_bytes['monero'] + data.pubkey
		return CoinAddr(
			proto = self.proto,
			addr = self.b58enc(step1 + self.keccak_256(step1).digest()[:4]))

	@check_data
	def to_viewkey(self, data):
		return self.proto.viewkey(data.viewkey_bytes.hex())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.ref: Reference file tests for the cmdtest.py test suite
"""

import os

from mmgen.util import capfirst
from mmgen.wallet import get_wallet_cls
from ..include.common import (
	imsg_r,
	ok,
	joinpath,
	cmp_or_die,
	ref_kafile_pass,
	read_from_file,
	sample_text
)
from .include.common import (
	dfl_words_file,
	ref_dir,
	chksum_pat,
	pwfile,
	ref_bw_file_spc,
	ref_enc_fn,
	cleanup_env,
	tool_enc_passwd,
	skip
)

from .base import CmdTestBase
from .shared import CmdTestShared

wpasswd = 'reference password'

class CmdTestRef(CmdTestBase, CmdTestShared):
	'saved reference address, password and transaction files'
	tmpdir_nums = [8]
	networks = ('btc', 'btc_tn', 'ltc', 'ltc_tn')
	passthru_opts = ('daemon_data_dir', 'rpc_port', 'coin', 'testnet')
	need_daemon = True
	sources = {
		'ref_addrfile':    '98831F3A{}[1,31-33,500-501,1010-1011]{}.addrs',
		'ref_segwitaddrfile':'98831F3A{}-S[1,31-33,500-501,1010-1011]{}.addrs',
		'ref_bech32addrfile':'98831F3A{}-B[1,31-33,500-501,1010-1011]{}.addrs',
		'ref_keyaddrfile': '98831F3A{}[1,31-33,500-501,1010-1011]{}.akeys.mmenc',
		'ref_viewkeyaddrfile': '98831F3A-XMR-M[1-3].vkeys',

		'ref_passwdfile_b32_24': '98831F3A-фубар@crypto.org-b32-24[1,4,1100].pws',
		'ref_passwdfile_b32_12': '98831F3A-фубар@crypto.org-b32-12[1,4,1100].pws',
		'ref_passwdfile_b58_10': '98831F3A-фубар@crypto.org-b58-10[1,4,1100].pws',
		'ref_passwdfile_b58_20': '98831F3A-фубар@crypto.org-b58-20[1,4,1100].pws',
		'ref_passwdfile_hex_32': '98831F3A-фубар@crypto.org-hex-32[1,4,1100].pws',
		'ref_passwdfile_hex_48': '98831F3A-фубар@crypto.org-hex-48[1,4,1100].pws',
		'ref_passwdfile_hex_64': '98831F3A-фубар@crypto.org-hex-64[1,4,1100].pws',
		'ref_passwdfile_bip39_12': '98831F3A-фубар@crypto.org-bip39-12[1,4,1100].pws',
		'ref_passwdfile_bip39_18': '98831F3A-фубар@crypto.org-bip39-18[1,4,1100].pws',
		'ref_passwdfile_bip39_24': '98831F3A-фубар@crypto.org-bip39-24[1,4,1100].pws',
		'ref_passwdfile_xmrseed_25': '98831F3A-фубар@crypto.org-xmrseed-25[1,4,1100].pws',
		'ref_passwdfile_hex2bip39_12': '98831F3A-фубар@crypto.org-hex2bip39-12[1,4,1100].pws',
		'ref_tx_file': { # data shared with ref_altcoin, autosign
			'btc': (
				'0B8D5A[15.31789,14,tl=1320969600].rawtx',
				'0C7115[15.86255,14,tl=1320969600].testnet.rawtx'
			),
			'ltc': (
				'AF3CDF-LTC[620.76194,1453,tl=1320969600].rawtx',
				'A5A1E0-LTC[1454.64322,1453,tl=1320969600].testnet.rawtx'
			),
			'bch': (
				'460D4D-BCH[10.19764,tl=1320969600].rawtx',
				'359FD5-BCH[6.68868,tl=1320969600].testnet.rawtx'
			),
			'eth': (
				'88FEFD-ETH[23.45495,40000].rawtx',
				'76CF8C-ETH[99.99895,50000].regtest.rawtx'
			),
			'mm1': (
				'5881D2-MM1[1.23456,50000].rawtx',
				'6BDB25-MM1[1.23456,50000].testnet.rawtx'
			),
			'etc': (
				'ED3848-ETC[1.2345,40000].rawtx',
				''
			)
		},
	}
	chk_data = {
		'ref_subwallet_sid': {
			'98831F3A:32L':'D66B4885',
			'98831F3A:1S':'20D95B09',
		},
		'ref_addrfile_chksum': {
			'btc': ('6FEF 6FB9 7B13 5D91', '424E 4326 CFFE 5F51'),
			'ltc': ('AD52 C3FE 8924 AAF0', '4EBE 2E85 E969 1B30'),
		},
		'ref_segwitaddrfile_chksum': {
			'btc': ('06C1 9C87 F25C 4EE6', '072C 8B07 2730 CB7A'),
			'ltc': ('63DF E42A 0827 21C3', '5DD1 D186 DBE1 59F2'),
		},
		'ref_bech32addrfile_chksum': {
			'btc': ('9D2A D4B6 5117 F02E', '0527 9C39 6C1B E39A'),
			'ltc': ('FF1C 7939 5967 AB82', 'ED3D 8AA4 BED4 0B40'),
		},
		'ref_keyaddrfile_chksum': {
			'btc': ('9F2D D781 1812 8BAD', '88CC 5120 9A91 22C2'),
			'ltc': ('B804 978A 8796 3ED4', '98B5 AC35 F334 0398'),
		},
		'ref_passwdfile_b32_12_chksum':       '7252 CD8D EF0D 3DB1',
		'ref_passwdfile_b32_24_chksum':       '8D56 3845 A072 A5B9',
		'ref_passwdfile_b58_10_chksum':       '534F CC1A 6701 9FED',
		'ref_passwdfile_b58_20_chksum':       'DDD9 44B0 CA28 183F',
		'ref_passwdfile_hex_32_chksum':       '05C7 3678 E25E BC32',
		'ref_passwdfile_hex_48_chksum':       '7DBB FFD0 633E DE6F',
		'ref_passwdfile_hex_64_chksum':       'F11D CB0A 8AE3 4D21',
		'ref_passwdfile_bip39_12_chksum':     'BF57 02A3 5229 CF18',
		'ref_passwdfile_bip39_18_chksum':     '31D3 1656 B7DC 27CF',
		'ref_passwdfile_bip39_24_chksum':     'E565 3A59 7D91 4671',
		'ref_passwdfile_xmrseed_25_chksum':   'B488 21D3 4539 968D',
		'ref_passwdfile_hex2bip39_12_chksum': '93AD 4AE2 03D1 8A0A',
	}
	cmd_group = ( # TODO: move to tooltest2
		('ref_words_to_subwallet_chk1',     'subwallet generation from reference words file (long subseed)'),
		('ref_words_to_subwallet_chk2',     'subwallet generation from reference words file (short subseed)'),
		('ref_subwallet_addrgen1',          'subwallet address file generation (long subseed)'),
		('ref_subwallet_addrgen2',          'subwallet address file generation (short subseed)'),
		('ref_subwallet_keygen1',           'subwallet key-address file generation (long subseed)'),
		('ref_subwallet_keygen2',           'subwallet key-address file generation (short subseed)'),
		('ref_addrfile_chk',                'saved reference address file'),
		('ref_segwitaddrfile_chk',          'saved reference address file (segwit)'),
		('ref_bech32addrfile_chk',          'saved reference address file (bech32)'),
		('ref_keyaddrfile_chk',             'saved reference key-address file'),

		('ref_passwdfile_chk_b58_20',       'saved reference password file (base58, 20 chars)'),
		('ref_passwdfile_chk_b58_10',       'saved reference password file (base58, 10 chars)'),
		('ref_passwdfile_chk_b32_24',       'saved reference password file (base32, 24 chars)'),
		('ref_passwdfile_chk_b32_12',       'saved reference password file (base32, 12 chars)'),
		('ref_passwdfile_chk_hex_32',       'saved reference password file (hexadecimal, 32 chars)'),
		('ref_passwdfile_chk_hex_48',       'saved reference password file (hexadecimal, 48 chars)'),
		('ref_passwdfile_chk_hex_64',       'saved reference password file (hexadecimal, 64 chars)'),
		('ref_passwdfile_chk_bip39_12',     'saved reference password file (BIP39, 12 words)'),
		('ref_passwdfile_chk_bip39_18',     'saved reference password file (BIP39, 18 words)'),
		('ref_passwdfile_chk_bip39_24',     'saved reference password file (BIP39, 24 words)'),
		('ref_passwdfile_chk_xmrseed_25',   'saved reference password file (Monero new-style mnemonic, 25 words)'),
		('ref_passwdfile_chk_hex2bip39_12', 'saved reference password file (hex-to-BIP39, 12 words)'),

#	Create the fake inputs:
#	('txcreate8',          'transaction creation (8)'),
		('ref_tx_chk',                   'signing saved reference tx file'),
		('ref_brain_chk_spc3',           'saved brainwallet (non-standard spacing)'),
		('ref_dieroll_chk_seedtruncate', 'saved dieroll wallet with extra entropy bits'),
		('ref_tool_decrypt',             'decryption of saved MMGen-encrypted file'),
	)

	@property
	def nw_desc(self):
		return '{} {}'.format(
			self.proto.coin,
			('Mainnet','Testnet')[self.proto.testnet])

	def _get_ref_subdir_by_coin(self,coin):
		return {
			'btc':  '',
			'bch':  '',
			'ltc':  'litecoin',
			'eth':  'ethereum',
			'etc':  'ethereum_classic',
			'xmr':  'monero',
			'zec':  'zcash',
			'dash': 'dash'
		}[coin.lower()]

	@property
	def ref_subdir(self):
		return self._get_ref_subdir_by_coin(self.proto.coin)

	def ref_words_to_subwallet_chk1(self):
		return self.ref_words_to_subwallet_chk('32L')

	def ref_words_to_subwallet_chk2(self):
		return self.ref_words_to_subwallet_chk('1S')

	def ref_words_to_subwallet_chk(self, ss_idx):
		wf = dfl_words_file
		ocls = get_wallet_cls('words')
		args = ['-d', self.tr.trash_dir, '-o', ocls.fmt_codes[-1], wf, ss_idx]

		t = self.spawn(
			'mmgen-subwalletgen',
			self.testnet_opt + args,
			extra_desc       = '(generate subwallet)',
			no_passthru_opts = True)
		t.expect(f'Generating subseed {ss_idx}')
		chk_sid = self.chk_data['ref_subwallet_sid'][f'98831F3A:{ss_idx}']
		fn = t.written_to_file(capfirst(ocls.desc))
		assert chk_sid in fn, f'incorrect filename: {fn} (does not contain {chk_sid})'
		ok()

		t = self.spawn(
			'mmgen-walletchk',
			self.testnet_opt + [fn],
			extra_desc       = '(check subwallet)',
			no_passthru_opts = True)
		t.expect(r'Valid MMGen native mnemonic data for Seed ID ([0-9A-F]*)\b', regex=True)
		sid = t.p.match.group(1)
		assert sid == chk_sid, f'subseed ID {sid} does not match expected value {chk_sid}'
		return t

	def ref_subwallet_addrgen(self, ss_idx, target='addr'):
		wf = dfl_words_file
		args = ['-d', self.tr.trash_dir, '--subwallet='+ss_idx, wf, '1-10']
		t = self.spawn(f'mmgen-{target}gen', args)
		t.expect(f'Generating subseed {ss_idx}')
		chk_sid = self.chk_data['ref_subwallet_sid'][f'98831F3A:{ss_idx}']
		assert chk_sid == t.expect_getend('Checksum for .* data ', regex=True)[:8]
		if target == 'key':
			t.expect('Encrypt key list? (y/N): ', 'n')
		fn = t.written_to_file(('Addresses', 'Secret keys')[target=='key'])
		assert chk_sid in fn, f'incorrect filename: {fn} (does not contain {chk_sid})'
		return t

	def ref_subwallet_addrgen1(self):
		return self.ref_subwallet_addrgen('32L')

	def ref_subwallet_addrgen2(self):
		return self.ref_subwallet_addrgen('1S')

	def ref_subwallet_keygen1(self):
		return self.ref_subwallet_addrgen('32L', target='key')

	def ref_subwallet_keygen2(self):
		return self.ref_subwallet_addrgen('1S', target='key')

	def ref_addrfile_chk(
			self,
			ftype  = 'addr',
			coin   = None,
			subdir = None,
			pfx    = None,
			mmtype = None,
			id_key = None,
			pat    = None):

		pat = pat or f'{self.nw_desc}.*Legacy'
		af_key = f'ref_{ftype}file' + ('_' + id_key if id_key else '')
		af_fn = CmdTestRef.sources[af_key].format(pfx or self.altcoin_pfx, '' if coin else self.tn_ext)
		af = joinpath(ref_dir, (subdir or self.ref_subdir, '')[ftype=='passwd'], af_fn)
		coin_arg = [] if coin is None else ['--coin='+coin]
		tool_cmd = ftype.replace('segwit', '').replace('bech32', '')+'file_chksum'
		t = self.spawn('mmgen-tool', coin_arg + ['--verbose', '-p1', tool_cmd, af])
		if ftype == 'keyaddr':
			t.do_decrypt_ka_data(pw=ref_kafile_pass, have_yes_opt=True)
		chksum_key = '_'.join([af_key, 'chksum'] + ([coin.lower()] if coin else []) + ([mmtype] if mmtype else []))
		rc = self.chk_data[chksum_key]
		ref_chksum = rc if (ftype == 'passwd' or coin) else rc[self.proto.base_coin.lower()][self.proto.testnet]
		if pat:
			t.expect(pat, regex=True)
		t.expect(chksum_pat, regex=True)
		m = t.p.match.group(0)
		cmp_or_die(ref_chksum, m)
		return t

	def ref_segwitaddrfile_chk(self):
		if not 'S' in self.proto.mmtypes:
			return skip(f'not supported by {self.proto.cls_name} protocol')
		return self.ref_addrfile_chk(ftype='segwitaddr', pat=f'{self.nw_desc}.*Segwit')

	def ref_bech32addrfile_chk(self):
		if not 'B' in self.proto.mmtypes:
			return skip(f'not supported by {self.proto.cls_name} protocol')
		return self.ref_addrfile_chk(ftype='bech32addr', pat=f'{self.nw_desc}.*Bech32')

	def ref_keyaddrfile_chk(self):
		return self.ref_addrfile_chk(ftype='keyaddr')

	def ref_passwdfile_chk(self, key, pat):
		return self.ref_addrfile_chk(ftype='passwd', id_key=key, pat=pat)

	def ref_passwdfile_chk_b58_20(self):
		return self.ref_passwdfile_chk(key='b58_20', pat=r'Base58.*len.* 20\b')
	def ref_passwdfile_chk_b58_10(self):
		return self.ref_passwdfile_chk(key='b58_10', pat=r'Base58.*len.* 10\b')
	def ref_passwdfile_chk_b32_24(self):
		return self.ref_passwdfile_chk(key='b32_24', pat=r'Base32.*len.* 24\b')
	def ref_passwdfile_chk_b32_12(self):
		return self.ref_passwdfile_chk(key='b32_12', pat=r'Base32.*len.* 12\b')
	def ref_passwdfile_chk_hex_32(self):
		return self.ref_passwdfile_chk(key='hex_32', pat=r'Hexadec.*len.* 32\b')
	def ref_passwdfile_chk_hex_48(self):
		return self.ref_passwdfile_chk(key='hex_48', pat=r'Hexadec.*len.* 48\b')
	def ref_passwdfile_chk_hex_64(self):
		return self.ref_passwdfile_chk(key='hex_64', pat=r'Hexadec.*len.* 64\b')
	def ref_passwdfile_chk_bip39_12(self):
		return self.ref_passwdfile_chk(key='bip39_12', pat=r'BIP39.*len.* 12\b')
	def ref_passwdfile_chk_bip39_18(self):
		return self.ref_passwdfile_chk(key='bip39_18', pat=r'BIP39.*len.* 18\b')
	def ref_passwdfile_chk_bip39_24(self):
		return self.ref_passwdfile_chk(key='bip39_24', pat=r'BIP39.*len.* 24\b')
	def ref_passwdfile_chk_xmrseed_25(self):
		return self.ref_passwdfile_chk(key='xmrseed_25', pat=r'Mon.*len.* 25\b')
	def ref_passwdfile_chk_hex2bip39_12(self):
		return self.ref_passwdfile_chk(key='hex2bip39_12', pat=r'BIP39.*len.* 12\b')

	def ref_tx_chk(self):
		fn = self.sources['ref_tx_file'][self.coin][bool(self.tn_ext)]
		if not fn:
			return
		tf = joinpath(ref_dir, self.ref_subdir, fn)
		wf = dfl_words_file
		self.write_to_tmpfile(pwfile, wpasswd)
		return self.txsign(wf, tf, save=False, has_label=True, view='y')

	def ref_brain_chk_spc3(self):
		return self.ref_brain_chk(bw_file=ref_bw_file_spc)

	def ref_dieroll_chk_seedtruncate(self):
		wf = joinpath(ref_dir, 'overflow128.b6d')
		return self.walletchk(wf, sid='8EC6D4A2')

	def ref_tool_decrypt(self):
		f = joinpath(ref_dir, ref_enc_fn)
		dec_file = joinpath(self.tmpdir, 'famous.txt')
		t = self.spawn(
			'mmgen-tool',
			['-q', 'decrypt', f, 'outfile='+dec_file, 'hash_preset=1'],
			env = cleanup_env(self.cfg))
		t.passphrase('data', tool_enc_passwd)
		t.written_to_file('Decrypted data')
		dec_txt = read_from_file(dec_file)
		imsg_r(dec_txt)
		cmp_or_die(sample_text+'\n', dec_txt) # file adds a newline to sample_text
		return t



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.new: Bitcoin new transaction class
"""

from ....tx.new import New as TxNew
from ....obj import MMGenTxID
from ....util import msg, fmt, make_chksum_6, die, suf
from ....color import pink
from .base import Base

class New(Base, TxNew):
	usr_fee_prompt = 'Enter transaction fee: '
	fee_fail_fs = 'Network fee estimation for {c} confirmations failed ({t})'
	no_chg_msg = 'Warning: Change address will be deleted as transaction produces no change'
	msg_insufficient_funds = 'Selected outputs insufficient to fund this transaction ({} {} needed)'

	async def set_gas(self, *, to_addr=None, force=False):
		return None

	def process_data_output_arg(self, arg):
		if any(arg.startswith(pfx) for pfx in ('data:', 'hexdata:')):
			if hasattr(self, '_have_op_return_data'):
				die(1, 'Transaction may have at most one OP_RETURN data output!')
			self._have_op_return_data = True
			from .op_return_data import OpReturnData
			OpReturnData(self.proto, arg) # test data for validity
			return arg

	@property
	def relay_fee(self):
		kb_fee = self.proto.coin_amt(self.rpc.cached['networkinfo']['relayfee'])
		ret = kb_fee * self.estimate_size() / 1024
		self.cfg._util.vmsg(f'Relay fee: {kb_fee} {self.coin}/kB, for transaction: {ret} {self.coin}')
		return ret

	@property
	def network_estimated_fee_label(self):
		return 'Network-estimated ({}, {} conf{})'.format(
			self.cfg.fee_estimate_mode.upper(),
			pink(str(self.cfg.fee_estimate_confs)),
			suf(self.cfg.fee_estimate_confs))

	def warn_fee_estimate_fail(self, fe_type):
		if not hasattr(self, '_fee_estimate_fail_warning_shown'):
			msg(self.fee_fail_fs.format(
				c = self.cfg.fee_estimate_confs,
				t = fe_type))
			self._fee_estimate_fail_warning_shown = True

	def network_fee_to_unit_disp(self, net_fee):
		return '{} sat/byte'.format(net_fee.fee.to_unit('satoshi') // 1024)

	async def get_rel_fee_from_network(self):
		try:
			ret = await self.rpc.call(
				'estimatesmartfee',
				self.cfg.fee_estimate_confs,
				self.cfg.fee_estimate_mode.upper())
			fee_per_kb = self.proto.coin_amt(ret['feerate']) if 'feerate' in ret else None
			fe_type = 'estimatesmartfee'
		except:
			args = self.rpc.daemon.estimatefee_args(self.rpc)
			ret = await self.rpc.call('estimatefee', *args)
			fee_per_kb = self.proto.coin_amt(ret)
			fe_type = 'estimatefee'

		if fee_per_kb is None:
			self.warn_fee_estimate_fail(fe_type)

		return self._net_fee(fee_per_kb, fe_type)

	# given tx size, rel fee and units, return absolute fee
	def fee_rel2abs(self, tx_size, amt_in_units, unit):
		return self.proto.coin_amt(int(amt_in_units * tx_size), from_unit=unit)

	# given network fee estimate in BTC/kB, return absolute fee using estimated tx size
	def fee_est2abs(self, net_fee):
		tx_size = self.estimate_size()
		ret = self.proto.coin_amt('1') * (net_fee.fee * self.cfg.fee_adjust * tx_size / 1024)
		if self.cfg.verbose:
			msg(fmt(f"""
				{net_fee.type.upper()} fee for {self.cfg.fee_estimate_confs} confirmations: {net_fee.fee} {self.coin}/kB
				TX size (estimated): {tx_size} bytes
				Fee adjustment factor: {self.cfg.fee_adjust:.2f}
				Absolute fee (net_fee.fee * adj_factor * tx_size / 1024): {ret} {self.coin}
			""").strip())
		return ret

	def convert_and_check_fee(self, fee, desc):
		abs_fee = self.feespec2abs(fee, self.estimate_size())
		if abs_fee is None:
			raise ValueError(f'{fee}: cannot convert {self.rel_fee_desc} to {self.coin}'
								+ ' because transaction size is unknown')
		if abs_fee is False:
			err = f'{fee!r}: invalid TX fee (not a {self.coin} amount or {self.rel_fee_desc} specification)'
		elif abs_fee > self.proto.max_tx_fee:
			err = f'{abs_fee} {self.coin}: {desc} fee too large (maximum fee: {self.proto.max_tx_fee} {self.coin})'
		elif abs_fee < self.relay_fee:
			err = f'{abs_fee} {self.coin}: {desc} fee too small (less than relay fee of {self.relay_fee} {self.coin})'
		else:
			return abs_fee
		msg(err)
		return False

	async def get_input_addrs_from_inputs_opt(self):
		# Bitcoin full node, call doesn't go to the network, so just call listunspent with addrs=[]
		return []

	def update_change_output(self, funds_left):
		if funds_left == 0: # TODO: test
			msg(self.no_chg_msg)
			self.outputs.pop(self.chg_idx)
		else:
			self.update_output_amt(self.chg_idx, funds_left)

	def check_fee(self):
		fee = self.sum_inputs() - self.sum_outputs()
		if fee > self.proto.max_tx_fee:
			c = self.proto.coin
			die('MaxFeeExceeded', f'Transaction fee of {fee} {c} too high! (> {self.proto.max_tx_fee} {c})')

	def final_inputs_ok_msg(self, funds_left):
		return 'Transaction produces {} {} in change'.format(funds_left.hl(), self.coin)

	def check_chg_addr_is_wallet_addr(self):
		if len(self.nondata_outputs) > 1 and not self.chg_output.mmid:
			self._non_wallet_addr_confirm('Change address is not an MMGen wallet address!')

	async def create_serialized(self, *, locktime=None):

		if not self.is_bump:
			# Set all sequence numbers to the same value, in conformity with the behavior of most modern wallets:
			do_rbf = self.proto.cap('rbf') and not self.cfg.no_rbf
			seqnum_val = self.proto.max_int - (2 if do_rbf else 1 if locktime else 0)
			for i in self.inputs:
				i.sequence = seqnum_val

		if not self.is_swap:
			self.inputs.sort_bip69()
			self.outputs.sort_bip69()

		inputs_list = [{
				'txid':     e.txid,
				'vout':     e.vout,
				'sequence': e.sequence
			} for e in self.inputs]

		outputs_dict = dict((e.addr, e.amt) if e.addr else ('data', e.data.hex()) for e in self.outputs)

		ret = await self.rpc.call('createrawtransaction', inputs_list, outputs_dict)

		if locktime and not self.is_bump:
			msg(f'Setting nLockTime to {self.info.strfmt_locktime(locktime)}!')
			assert isinstance(locktime, int), 'locktime value not an integer'
			self.locktime = locktime
			ret = ret[:-8] + bytes.fromhex(f'{locktime:08x}')[::-1].hex()

		# TxID is set only once!
		self.txid = MMGenTxID(make_chksum_6(bytes.fromhex(ret)).upper())

		self.update_serialized(ret)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.httpd: WSGI http server
"""

from wsgiref.simple_server import make_server, WSGIRequestHandler

from mmgen.util import msg
from mmgen.util2 import port_in_use

class SilentRequestHandler(WSGIRequestHandler):

	def log_request(self, code='-', size='-'):
		return None

class HTTPD:

	def start(self):

		if port_in_use(self.port):
			msg(f'\nPort {self.port} in use. Assuming {self.name} is running')
			return True

		self.httpd = make_server(
			'localhost',
			self.port,
			self.application,
			handler_class = SilentRequestHandler)

		import threading
		t = threading.Thread(target=self.httpd.serve_forever, name=f'{type(self).__name__} thread')
		t.daemon = True
		t.start()

	def stop(self):
		self.httpd.server_close()

	def application(self, environ, start_response):

		method = environ['REQUEST_METHOD']

		response_body = self.make_response_body(method, environ)

		status = '200 OK'
		response_headers = [
			('Content-Type', self.content_type),
			('Content-Length', str(len(response_body)))
		]

		start_response(status, response_headers)

		return [response_body]



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# ------------------------------------------------------------------------------

# MMGen note: this is a sorted version of the wordlist.
# The unsorted original can be found at:
#    https://github.com/spesmilo/electrum/blob/1.9.5/lib/mnemonic.py
# Electrum - lightweight Bitcoin client. Copyright (C) 2011 thomasv@gitorious

# Also available at:
#    https://github.com/monero-project/monero/blob/master/src/mnemonics/english_old.h

words = tuple("""
able
about
above
abuse
accept
accident
ache
acid
across
act
action
actually
add
admire
admit
adore
advice
affection
afraid
after
afternoon
again
against
age
agony
agree
ahead
aim
air
alas
alive
allow
almost
alone
along
already
also
alter
although
always
amaze
among
ancient
angel
anger
angry
animal
annoy
another
answer
any
anymore
anyone
anything
anyway
anywhere
apart
apologize
appear
apple
appreciate
approach
army
around
arrive
arrow
art
ash
ashamed
aside
ask
asleep
attack
attempt
attention
aunt
autumn
avoid
await
awake
awaken
away
awe
awkward
baby
back
bag
balance
ball
bang
bank
bar
bare
barely
bathroom
battle
beam
beard
beaten
beautiful
beauty
became
become
bedroom
bee
been
before
beg
began
begin
begun
behind
belief
believe
belly
belong
bench
beneath
beside
best
better
between
beyond
bid
bird
birth
birthday
bit
bite
bitter
black
blade
blame
blank
blanket
bleed
blend
bless
blind
blink
bliss
block
blonde
blood
bloody
bloom
blossom
blow
blue
blur
blush
board
boat
body
bold
bomb
bond
bone
book
boom
born
both
bother
bottle
bottom
bought
bounce
bound
bowl
box
boyfriend
brain
branch
brand
brave
bread
break
breast
breath
breathe
breeze
bridge
bright
bring
broke
broken
brother
brown
bruise
brush
bubble
build
built
bullet
bump
burden
burn
burst
bury
bus
busy
butterfly
button
buy
cage
calm
came
candle
canvas
cap
capture
card
careful
carefully
caress
carry
carve
cast
catch
caught
cause
cease
ceiling
cell
center
certain
certainly
chain
chair
chance
change
character
charm
chase
cheap
cheat
check
cheek
cheer
cheese
chest
child
childhood
children
chill
chin
chocolate
choice
choke
choose
church
cigarette
circle
city
claim
class
claw
clay
clean
clear
clearly
click
climb
cling
clock
close
closet
clothes
cloud
clue
clutch
coat
coffee
cold
collapse
collect
college
color
comfort
commit
common
companion
company
compare
complain
complete
completely
concern
concrete
confidence
confuse
confusion
connect
connection
consider
constant
constantly
consume
content
continue
control
conversation
cookie
cool
core
corner
cost
couch
count
country
couple
courage
course
cousin
cover
coward
crack
crap
crash
crave
crawl
crazy
creak
cream
create
creation
creator
creature
creek
creep
crime
crimson
cross
crowd
cruel
crumble
crush
cry
crystal
cup
curl
curse
curtain
curve
cute
dad
daddy
daily
damn
dance
danger
dark
darkness
darling
dart
date
daughter
dawn
deadly
deal
dear
death
decay
decide
decision
deep
deeply
defeat
defense
define
delicate
delight
demand
demon
deny
depend
depress
depression
depth
descend
describe
desert
deserve
desire
desk
despair
desperate
desperately
despite
destination
destroy
determine
devil
dew
diamond
difference
different
dig
dim
dinner
direction
dirt
dirty
disappear
discover
disguise
dish
distance
distant
doctor
dog
doll
doom
door
dot
double
doubt
dove
down
drag
dragon
drama
draw
drawn
dread
dream
dreamer
dress
drift
drink
drive
driver
drop
drove
drug
drum
drunk
dry
duck
dude
due
dull
dumb
during
dust
dusty
dwell
each
eager
early
earth
easily
easy
eat
echo
edge
effort
egg
either
embarrass
embrace
emotion
emptiness
empty
end
endless
enemy
energy
engine
enjoy
enough
enter
entire
erase
escape
especially
essence
eternal
eternity
even
evening
eventually
ever
everybody
everyday
everyone
everything
everytime
everywhere
evil
exactly
example
except
excuse
exist
existence
expect
experience
explain
explode
explore
express
expression
eye
fact
fade
fail
faint
fairy
faith
false
familiar
family
fan
fantasy
fate
father
fault
favorite
fear
feast
feather
feed
feet
felicity
few
field
figure
fill
final
finally
find
finger
finish
fire
first
fish
fist
five
fix
flag
flame
flare
flash
flat
flesh
flicker
flight
flip
flirt
float
flood
floor
flow
flower
flutter
fly
focus
foe
fog
fold
follow
foot
football
footstep
force
forehead
forest
forever
forget
forgive
forgot
forgotten
form
former
forth
forward
fought
foul
fragile
frame
freak
free
freedom
freely
freeze
fresh
friend
friendship
fright
frighten
front
frost
frown
frozen
fruit
frustrate
fully
funny
fur
further
future
gain
game
garden
gas
gasp
gather
gay
gaze
generation
gentle
gently
ghost
giant
gift
giggle
girl
girlfriend
give
given
glad
glance
glare
glass
glide
glorious
glory
glove
glow
goal
god
gone
good
goodbye
goose
got
gotta
gotten
government
grab
grace
grade
grand
grandma
grant
grasp
grass
great
green
greet
grew
grey
grief
grin
grip
ground
group
grow
grown
guard
guess
guide
guilt
guilty
guitar
gun
guy
hair
half
hallway
hand
handle
hang
happen
happiness
happy
hardly
harm
harmony
harsh
hate
hatred
haunt
heal
health
heard
heart
heartbeat
heat
heaven
heavy
heel
held
hello
help
here
hero
hey
hidden
hide
high
hill
history
hit
hollow
holy
home
honest
honey
honor
hook
hop
hope
hopefully
horizon
horrible
horse
hospital
hour
house
however
howl
hug
huge
hum
human
hundred
hunger
hungry
hunt
hurl
hurry
hurt
husband
ice
idea
idiot
ignorance
ignore
illuminate
illusion
image
imagination
imagine
important
impossible
inch
indeed
inhale
ink
inner
innocence
innocent
insane
inside
instead
insult
interest
into
invisible
invite
iron
itself
jaw
jealous
jeans
job
jock
join
joke
journey
joy
judge
juice
jump
just
keep
kept
key
kick
kid
kill
king
kingdom
kiss
kitchen
knee
knife
knock
know
knowledge
known
lace
lady
laid
land
language
large
laugh
laughter
law
lazy
lead
leaf
leap
learn
least
leave
led
left
leg
less
lesson
letter
level
lick
lifeless
lift
lightning
like
limb
limit
line
linger
lip
liquid
listen
little
lonely
long
look
loose
lord
loser
loss
lot
loud
love
lovely
lucky
lunch
lung
lust
machine
made
magic
make
mama
manage
many
march
mark
marry
mask
mass
master
match
math
matter
maybe
mean
meant
measure
meet
melody
melt
memory
men
mend
mention
mercy
mere
mess
message
metal
middle
midnight
might
mighty
milk
million
mind
mine
minute
mirror
misery
mist
mistake
mix
moan
mock
mold
mom
moment
money
monkey
monster
month
moon
moonlight
moral
more
morning
mostly
mother
motion
mountain
mourn
mouth
movement
movie
much
mud
mumble
murder
muscle
muse
music
mutter
myself
mystery
nail
naked
name
nation
natural
nature
near
nearly
neck
need
needle
neighbor
neither
nerd
nerve
nervous
never
new
next
nice
night
nightmare
nine
nobody
nod
noise
normal
north
nose
note
nothing
notice
nowhere
numb
number
object
observe
obviously
ocean
odd
ode
off
offer
often
okay
once
only
open
opinion
opposite
orange
order
other
ourselves
out
outside
over
own
pack
page
painful
paint
pair
pale
palm
pants
paper
paradise
parent
park
party
pass
passion
past
path
pathetic
patience
patient
pattern
pause
pay
peace
peaceful
peach
peel
peer
pen
pencil
people
perfect
perfection
perfectly
perhaps
person
petal
phone
physical
picture
pie
piece
pierce
pig
pile
pillow
pink
pity
place
plain
plan
planet
plant
plastic
plate
play
please
pleasure
pocket
poem
poet
poetry
point
poison
ponder
pool
poor
pop
position
possess
possible
possibly
pound
pour
power
powerful
practice
praise
pray
prayer
precious
prefer
prepare
presence
present
press
pressure
pretend
pretty
price
pride
prince
princess
probably
problem
process
promise
protect
proud
prove
puff
pull
pulse
punch
pure
purple
purpose
push
put
puzzle
queen
question
quick
quickly
quiet
quietly
quit
quite
rabbit
radio
rainbow
raise
random
rant
rape
rare
rather
raw
reach
ready
real
reality
realize
really
reason
rebel
recall
receive
red
reflect
reflection
refuse
regret
relationship
relax
release
remain
remember
remind
remove
repeat
replace
reply
respect
respond
rest
retreat
return
reveal
revenge
rhyme
rhythm
rich
ring
rip
ripple
rise
river
road
roam
roar
rock
roll
root
rope
rose
rough
royal
rub
rude
ruin
rule
run
rush
sad
sadness
safe
said
sail
salt
salty
sanctuary
sane
sanity
sat
satisfy
save
savior
scale
scar
scary
scatter
scene
scent
school
score
scrape
scratch
scream
screen
screw
sea
search
season
second
secret
seek
seem
self
selfish
sense
sent
sentence
separate
serious
serve
settle
seven
several
shade
shadow
shall
shame
shape
sharp
shatter
sheep
sheet
shelter
shield
shift
shimmer
shine
shiny
ship
shirt
shiver
shock
shoe
shook
shoot
shore
should
shoulder
shout
shove
show
shower
shown
shut
shy
sick
sigh
sign
silence
silent
silently
silly
silver
simple
simply
since
single
sink
sister
situation
six
size
skill
skin
skip
sky
slam
slap
sleep
slept
slice
slide
slight
slightly
slowly
small
smart
smell
smile
smoke
smooth
snake
snap
sneak
snow
soak
soar
sob
social
society
soft
softly
soldier
some
somebody
someday
somehow
someone
something
sometimes
somewhere
son
song
soon
soothe
sorrow
sorry
sort
soul
sound
space
speak
special
speed
spell
spend
spider
spill
spin
spine
spiral
spirit
split
spoken
sport
spot
spread
spring
square
squeeze
stab
stage
stain
stair
stand
stare
start
state
stay
steady
steal
steel
step
stick
still
sting
stir
stock
stole
stolen
stomach
stone
stood
stop
storm
story
straight
strain
strange
stranger
stray
stream
street
strength
stress
stretch
strife
strike
string
strip
stroke
strong
struggle
stubborn
stuck
student
study
stuff
stumble
stun
stupid
style
subject
such
sudden
suddenly
suffer
suffocate
suicide
suit
summer
sunlight
sunset
support
suppose
sure
surely
surface
surprise
surround
survive
suspend
swallow
sway
swear
sweat
sweet
swell
swim
swing
swirl
sword
sympathy
table
taint
take
taken
talent
talk
tangle
tap
task
taste
taught
tea
teach
teacher
team
tease
teeth
tell
tender
terrible
terror
test
thank
them
themselves
then
there
thick
thigh
thing
think
third
thorn
those
thought
thousand
thread
threaten
three
threw
throat
throne
through
throughout
throw
thrown
thump
thunder
tickle
tide
tie
tight
tightly
time
tiny
tired
today
toe
together
tomorrow
tongue
tonight
too
tool
torment
torture
toss
total
touch
tough
toward
tower
town
trace
trade
trail
train
travel
treasure
treat
tree
tremble
trick
trickle
trip
trouble
truck
true
truly
trust
truth
try
tuck
tumble
tune
turn
twenty
twice
twirl
twist
two
type
ugly
unable
uncle
under
underneath
understand
universe
unknown
unless
unlike
unseen
upon
upset
use
useless
usually
utter
vain
valley
value
vast
veil
vein
velvet
verse
very
victim
view
violence
vision
visit
voice
void
volume
waist
wait
walk
wall
wander
want
war
warm
warmth
warn
warrior
wash
waste
watch
water
wave
weak
weakness
weapon
weary
weather
weave
week
weep
weight
weird
welcome
wet
whatever
wheel
whenever
whether
whisper
whistle
white
whole
wife
wild
win
window
winter
wipe
wise
wish
witch
within
wolf
woman
women
won
wonder
wonderful
wood
work
world
worry
worse
worship
worst
worth
worthless
wound
wow
wrap
wrinkle
wrist
write
written
wrong
wrote
yard
yeah
yearn
yell
yellow
yesterday
yet
young
yours
yourself
youth
""".split())



#!/usr/bin/env python3

"""
test.modtest_d.amt: CoinAmt unit tests for the MMGen suite
"""

from decimal import Decimal

from mmgen.protocol import init_proto
from mmgen.tx.new import parse_fee_spec
from mmgen.cfg import Config
from mmgen.amt import TokenAmt

from ..include.common import cfg, vmsg

def get_protos(data):
	return {coin: init_proto(cfg, coin, need_amt=True) for coin in set(d[0] for d in data)}

def test_to_unit(data):
	protos = get_protos(data)
	for proto, amt, unit, chk in data:
		amt = protos[proto].coin_amt(amt)
		res = amt.to_unit(unit)
		vmsg(f'  {proto.upper()} {amt.fmt(8)} => {res:<14} {unit}')
		if '.' in chk:
			assert res == Decimal(chk), f'{res} != {Decimal(chk)}'
		else:
			assert res == int(chk), f'{res} != {int(chk)}'
	return True

def test_fee_spec(data):
	protos = get_protos(data)
	for proto, spec, amt, unit in data:
		vmsg(f'  {proto.upper():6} {spec:<5} => {amt:<4} {unit}')
		res = parse_fee_spec(protos[proto], spec)
		assert res.amt == amt, f'  {res.amt} != {amt}'
		assert res.unit == unit, f'  {res.unit} != {unit}'
	return True

class unit_tests:

	altcoin_deps = ('fee_spec_alt', 'to_unit_alt', 'token_amt')

	def to_unit(self, name, ut, desc='CoinAmt.to_unit() (BTC)'):
		return test_to_unit((
			('btc', '0.00000001', 'satoshi', '1'),
			('btc', '1.23456789', 'satoshi', '123456789')))

	def to_unit_alt(self, name, ut, desc='CoinAmt.to_unit() (LTC, BCH, ETH, XMR)'):
		return test_to_unit((
			('ltc', '0.00000001',           'satoshi', '1'),
			('ltc', '1.23456789',           'satoshi', '123456789'),
			('bch', '0.00000001',           'satoshi', '1'),
			('bch', '1.23456789',           'satoshi', '123456789'),
			('eth', '1.234567890123456789', 'wei',     '1234567890123456789'),
			('eth', '1.234567890123456789', 'Kwei',    '1234567890123456.789'),
			('eth', '1.234567890123456789', 'Mwei',    '1234567890123.456789'),
			('eth', '1.234567890123456789', 'finney',  '1234.567890123456789'),
			('eth', '0.000000012345678901', 'Mwei',    '12345.678901'),
			('eth', '0.000000000000000001', 'Kwei',    '0.001'),
			('eth', '0.000000000000000001', 'Gwei',    '0.000000001'),
			('eth', '0.00000001',           'Gwei',    '10'),
			('eth', '1',                    'Gwei',    '1000000000'),
			('eth', '1',                    'finney',  '1000'),
			('xmr', '1',                    'atomic',  '1000000000000'),
			('xmr', '0.000000000001',       'atomic',  '1'),
			('xmr', '1.234567890123',       'atomic',  '1234567890123')))

	def fee_spec(self, name, ut, desc='fee spec parsing (BTC)'):
		return test_fee_spec((
			('btc', '32s',  '32', 'satoshi'),
			('btc', '1s',   '1',  'satoshi')))

	def fee_spec_alt(self, name, ut, desc='fee spec parsing (LTC, BCH, ETH, XMR)'):
		return test_fee_spec((
			('ltc', '3.07s', '3.07', 'satoshi'),
			('bch', '3.07s', '3.07', 'satoshi'),
			('eth', '3.07G', '3.07', 'Gwei'),
			('eth', '37M',   '37',   'Mwei'),
			('eth', '3701w', '3701', 'wei'),
			('eth', '3.07M', '3.07', 'Mwei'),
			('xmr', '3.07a', '3.07', 'atomic')))

	def token_amt(self, name, ut, desc='TokenAmt (ETH)'):
		for n, dec, unit, chk in (
			(1234567,                    6,  'atomic', '1.234567'),
			('1.234567',                 6,  None,     '1.234567'),
			(1234567,                    22, 'atomic', '0.0000000000000001234567'),
			('0.0000000000000001234567', 22, None,     '0.0000000000000001234567'),
		):
			amt = TokenAmt(n, decimals=dec, from_unit=unit)
			amt_disp = amt.hl(color=False)
			vmsg('  ' + amt_disp)
			if unit == 'atomic':
				assert amt.to_unit('atomic') == n
			assert amt_disp == chk, f'{amt_disp} != {chk}'
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.msg: Ethereum base protocol message signing classes
"""

from ...msg import coin_msg

class coin_msg(coin_msg):

	include_pubhash = False
	sigdata_pfx = '0x'
	msghash_types = ('eth_sign', 'raw') # first-listed is the default

	class unsigned(coin_msg.unsigned):

		async def do_sign(self, wif, message, msghash_type):
			from .util import ec_sign_message_with_privkey
			return ec_sign_message_with_privkey(self.cfg, message, bytes.fromhex(wif), msghash_type)

	class signed_online(coin_msg.signed_online):

		async def do_verify(self, addr, sig, message, msghash_type):
			from ...tool.coin import tool_cmd
			from .util import ec_recover_pubkey
			return tool_cmd(
				self.cfg,
				proto = self.proto).pubhex2addr(
					ec_recover_pubkey(self.cfg, message, sig, msghash_type)) == addr

	class exported_sigs(coin_msg.exported_sigs, signed_online):
		pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.bal: Bitcoin base protocol tracking wallet balance class
"""

from ....tw.bal import TwGetBalance
from ....tw.shared import get_tw_label
from ....rpc import rpc_init

class BitcoinTwGetBalance(TwGetBalance):

	async def __init__(self, cfg, proto, *, minconf, quiet):
		self.rpc = await rpc_init(cfg, proto)
		self.walletinfo = await self.rpc.walletinfo
		await super().__init__(cfg, proto, minconf=minconf, quiet=quiet)

	start_labels = ('TOTAL', 'Non-MMGen', 'Non-wallet')
	conf_cols = {
		'unconfirmed': 'Unconfirmed',
		'lt_minconf':  '<{minconf} confs',
		'ge_minconf':  '>={minconf} confs',
	}

	async def create_data(self):
		lbl_id = ('account', 'label')['label_api' in self.rpc.caps]
		for d in await self.rpc.call('listunspent', 0):
			tw_lbl = get_tw_label(self.proto, d[lbl_id])
			if tw_lbl:
				if tw_lbl.mmid.type == 'mmgen':
					label = tw_lbl.mmid.obj.sid
					if label not in self.data:
						self.data[label] = self.balance_info()
				else:
					label = 'Non-MMGen'
			else:
				label = 'Non-wallet'

			amt = self.proto.coin_amt(d['amount'])

			if not d['confirmations']:
				self.data['TOTAL']['unconfirmed'] += amt
				self.data[label]['unconfirmed'] += amt

			col_key = ('lt_minconf', 'ge_minconf')[d['confirmations'] >= self.minconf]
			self.data['TOTAL'][col_key] += amt
			self.data[label][col_key] += amt

			if d['spendable']: # TODO: use 'solvable' for descriptor wallets?
				self.data[label]['spendable'] += amt

	def format(self, color):

		def gen_spendable_warning():
			if check_spendable:
				for k, v in self.data.items():
					if v['spendable']:
						yield red(f'Warning: this wallet contains PRIVATE KEYS for {k} outputs!')

		if color:
			from ....color import red
		else:
			from ....color import nocolor as red

		desc_wallet = self.walletinfo.get('descriptors')
		check_spendable = not desc_wallet or (desc_wallet and self.walletinfo['private_keys_enabled'])

		warning = '\n'.join(gen_spendable_warning())

		return super().format(color) + ('\n' if warning else '') + warning



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.rune: THORChain RUNE tests for the cmdtest.py test suite
"""

from hashlib import md5

from mmgen.fileutil import get_data_from_file

from .include.common import dfl_sid, dfl_words_file
from .include.proxy import TestProxy
from .httpd.thornode.rpc import ThornodeRPCServer
from .ethdev import CmdTestEthdevMethods
from .base import CmdTestBase
from .shared import CmdTestShared
from .swap import CmdTestSwapMethods

class CmdTestRune(CmdTestEthdevMethods, CmdTestBase, CmdTestShared):
	'THORChain RUNE tracking wallet and transacting operations'
	networks = ('rune',)
	passthru_opts = ('coin', 'http_timeout')
	tmpdir_nums = [50]
	color = True
	menu_prompt = 'efresh balance:\b'
	txhex_chksum = '83f85785'

	cmd_group_in = (
		('subgroup.init',   []),
		('subgroup.main',   ['init']),
		('rpc_server_stop', 'stopping the Thornode RPC server'),
	)
	cmd_subgroups = {
		'init': (
			'initializing wallets',
			('addrgen',    'generating addresses'),
			('addrimport', 'importing addresses'),
		),
		'main': (
			'tracking wallet and transaction operations',
			('twview',               'viewing unspent outputs in tracking wallet'),
			('bal_refresh',          'refreshing address balance in tracking wallet'),
			('txcreate1',            'creating a transaction'),
			('txsign1',              'signing the transaction'),
			('txsend1_test',         'testing whether the transaction can be sent'),
			('txsend1',              'sending the transaction'),
			('txhex1',               'dumping the transaction hex'),
		),
	}

	def __init__(self, cfg, trunner, cfgs, spawn):
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)
		if trunner is None:
			return

		self.eth_opts = [f'--outdir={self.tmpdir}', '--regtest=1', '--quiet']
		self.eth_opts_noquiet = [f'--outdir={self.tmpdir}', '--regtest=1']

		self.rune_opts = self.eth_opts

		from mmgen.protocol import init_proto
		self.proto = init_proto(cfg, network_id=self.proto.coin + '_rt', need_amt=True)
		self.spawn_env['MMGEN_BOGUS_SEND'] = ''

		self.rpc_server = ThornodeRPCServer()
		self.rpc_server.start()

		TestProxy(self, cfg)

		self.txhex_file = f'{self.tmpdir}/tx_dump.hex'

	def addrgen(self):
		return self._addrgen()

	def addrimport(self):
		return self._addrimport()

	def twview(self):
		return self.spawn('mmgen-tool', self.rune_opts + ['twview'])

	def bal_refresh(self):
		t = self.spawn('mmgen-tool', self.rune_opts + ['listaddresses', 'interactive=1'])
		t.expect(self.menu_prompt, 'R')
		t.expect('menu): ', '3\n')
		t.expect('(y/N): ', 'y')
		t.expect(r'Total RUNE: \S*\D9876.54321321\D', regex=True)
		t.expect('address #3 refreshed')
		t.expect(self.menu_prompt, 'q')
		return t

	def txcreate1(self):
		t = self.spawn('mmgen-txcreate', self.rune_opts + ['98831F3A:X:2,54.321'])
		t.expect(self.menu_prompt, 'q')
		t.expect('spend from: ', '3\n')
		t.expect('(y/N): ', 'y') # add comment?
		t.expect('Comment: ', 'RUNE Boy\n')
		t.expect('view: ', 'y')
		t.expect('to continue: ', 'z')
		t.expect('(y/N): ', 'y') # save?
		t.written_to_file('Unsigned transaction')
		return t

	def txsign1(self):
		return self.txsign_ui_common(
			self.spawn(
				'mmgen-txsign',
				self.rune_opts + [self.get_file_with_ext('rawtx'), dfl_words_file],
				no_passthru_opts = ['coin']),
			has_label = True)

	def txsend1_test(self):
		return self._txsend(add_opts=['--test', f'--proxy=localhost:{TestProxy.port}'], test=True)

	def txsend1(self):
		return self._txsend()

	def _txsend(self, add_opts=[], *, test=False, dump_hex=False):
		t = self.spawn(
			'mmgen-txsend',
			self.rune_opts + add_opts + [self.get_file_with_ext('sigtx')],
			no_passthru_opts = ['coin'])
		t.expect('view: ', 'y')
		t.expect('to continue: ', 'z')
		t.expect('(y/N): ', 'n') # edit comment?
		if dump_hex:
			t.written_to_file('hex data')
		elif test:
			t.expect('can be sent')
		else:
			t.expect('to confirm: ', 'YES\n')
			t.written_to_file('Sent transaction')
		return t

	def txhex1(self):
		t = self._txsend(add_opts=[f'--dump-hex={self.txhex_file}'], dump_hex=True)
		t.read()
		txhex = get_data_from_file(self.cfg, self.txhex_file, silent=True)
		assert md5(txhex.encode()).hexdigest()[:8] == self.txhex_chksum
		return t

	def rpc_server_stop(self):
		return CmdTestSwapMethods._thornode_server_stop(
			self, attrname='rpc_server', name='Thornode RPC server')
